# ---------------------------------------------------------------
# 这个脚本向你展示了如何使用 onnxruntime 对 PPQ 导出的模型进行推理
# 你需要注意，Onnxruntime 可以运行各种各样的量化方案，但模型量化对 Onnxruntime 而言几乎无法起到加速作用
# 你可以使用 Onnxruntime 来验证量化方案以及 ppq 量化的正确性，但这不是一个合理的部署平台
# 修改 QUANT_PLATFROM 来使用不同的量化方案。

# This Script export ppq internal graph to onnxruntime,
# you should notice that onnx is designed as an Open Neural Network Exchange format.
# It has the capbility to describe most of ppq's quantization policy including combinations of:
#   Symmtrical, Asymmtrical, POT, Per-channel, Per-Layer
# However onnxruntime can not accelerate quantized model in most cases,
# you are supposed to use onnxruntime for verifying your network quantization result only.
# ---------------------------------------------------------------

# For this onnx inference test, all test data is randomly picked.
# If you want to use real data, just rewrite the defination of SAMPLES
import onnxruntime
import torch
from tqdm import tqdm

from esp_ppq import *
from esp_ppq.api import *

QUANT_PLATFROM = TargetPlatform.ONNXRUNTIME
MODEL = 'model.onnx'
INPUT_SHAPE = [1, 3, 224, 224]
SAMPLES = [torch.rand(size=INPUT_SHAPE) for _ in range(256)]  # rewirte this to use real data.
DEVICE = 'cuda'
FINETUNE = False
QS = QuantizationSettingFactory.default_setting()
EXECUTING_DEVICE = 'cuda'
REQUIRE_ANALYZE = True

# -------------------------------------------------------------------
# 下面向你展示了常用参数调节选项：
# -------------------------------------------------------------------
QS.lsq_optimization = FINETUNE  # 启动网络再训练过程，降低量化误差
QS.lsq_optimization_setting.steps = 500  # 再训练步数，影响训练时间，500 步大概几分钟
QS.lsq_optimization_setting.collecting_device = 'cuda'  # 缓存数据放在那，cuda 就是放在 gpu，如果显存超了你就换成 'cpu'

print('正准备量化你的网络，检查下列设置:')
print(f'TARGET PLATFORM      : {QUANT_PLATFROM.name}')
print(f'NETWORK INPUTSHAPE   : {INPUT_SHAPE}')

# ENABLE CUDA KERNEL 会加速量化效率 3x ~ 10x，但是你如果没有装相应编译环境的话是编译不了的
# 你可以尝试安装编译环境，或者在不启动 CUDA KERNEL 的情况下完成量化：移除 with ENABLE_CUDA_KERNEL(): 即可
with ENABLE_CUDA_KERNEL():
    qir = quantize_onnx_model(
        onnx_import_file=MODEL,
        calib_dataloader=SAMPLES,
        calib_steps=128,
        setting=QS,
        input_shape=INPUT_SHAPE,
        collate_fn=lambda x: x.to(EXECUTING_DEVICE),
        platform=QUANT_PLATFROM,
        do_quantize=True,
    )

    # -------------------------------------------------------------------
    # PPQ 计算量化误差时，使用信噪比的倒数作为指标，即噪声能量 / 信号能量
    # 量化误差 0.1 表示在整体信号中，量化噪声的能量约为 10%
    # 你应当注意，在 graphwise_error_analyse 分析中，我们衡量的是累计误差
    # 网络的最后一层往往都具有较大的累计误差，这些误差是其前面的所有层所共同造成的
    # 你需要使用 layerwise_error_analyse 逐层分析误差的来源
    # -------------------------------------------------------------------
    print('正计算网络量化误差(SNR)，最后一层的误差应小于 0.1 以保证量化精度:')
    reports = graphwise_error_analyse(
        graph=qir,
        running_device=EXECUTING_DEVICE,
        steps=32,
        dataloader=SAMPLES,
        collate_fn=lambda x: x.to(EXECUTING_DEVICE),
    )
    for op, snr in reports.items():
        if snr > 0.1:
            ppq_warning(f'层 {op} 的累计量化误差显著，请考虑进行优化')

    if REQUIRE_ANALYZE:
        print('正计算逐层量化误差(SNR)，每一层的独立量化误差应小于 0.1 以保证量化精度:')
        layerwise_error_analyse(
            graph=qir,
            running_device=EXECUTING_DEVICE,
            interested_outputs=None,
            dataloader=SAMPLES,
            collate_fn=lambda x: x.to(EXECUTING_DEVICE),
        )

    print('网络量化结束，正在生成目标文件:')
    export_ppq_graph(graph=qir, platform=QUANT_PLATFROM, graph_save_to='quantized.onnx')



import torch

from esp_ppq.core import (
    OBSERVER_ISOTONE_OBSERVER_AXIS,
    QuantizationProperty,
    QuantizationStates,
    TensorQuantizationConfig,
    ppq_warning,
)
from esp_ppq.IR.base.graph import Variable

from .base import BaseTensorObserver
from .range import minmax_to_scale_offset


class TorchIsotoneObserver(BaseTensorObserver):
    def __init__(self, watch_on: Variable, quant_cfg: TensorQuantizationConfig):
        super().__init__(watch_on, quant_cfg)
        self._cache = []
        if OBSERVER_ISOTONE_OBSERVER_AXIS not in quant_cfg.detail:
            ppq_warning('Initializing Torch Isotone Observer with implicit axis is not recommended.')
            self.axis = -1
        else:
            self.axis = quant_cfg.detail[OBSERVER_ISOTONE_OBSERVER_AXIS]

    """For softmax or sigmoid activations, usually we just need
    argmax(softmax(x)) == argmax(softmax(quant(x)))

    Inspired by this Property, Isotone Observer is designed to provide an order-preserving calibration method,
        which cares only about argmax(x) [or argmin(x)]

    To keep argmax(x) == argmax(quant(x)), we only need to
        distinguish the largest element and the second largert element with quantization

        let L1 represents the largest element of x,
        while L2 represents the second largest.

        For symmetric quantization:
        We want
            1. L1 - L2 > scale
            2. round(L2 / scale) <= (quant_max - .5)
    Args:
        BaseTensorObserver ([type]): [description]
    """

    def observe(self, value: torch.Tensor):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        assert isinstance(value, torch.Tensor), 'IsotoneObserver can only deal with torch Tensor values'
        assert value.numel() > 0, f'You are observing an empty tensor({self._watch_on.name}).'
        if self._quant_cfg.state == QuantizationStates.INITIAL:
            if self._quant_cfg.policy.has_property(QuantizationProperty.PER_TENSOR):
                # flatten value as [-1, num_of_elements in isotone axis]
                if value.ndim > 1:
                    value = value.transpose(dim0=self.axis, dim1=-1)
                    value = value.flatten(start_dim=0, end_dim=-2)
                value, _ = torch.topk(value, k=2, dim=self.axis, largest=True, sorted=True)
                if value.ndim <= 1:
                    value = value.unsqueeze(0)
                self._cache.append(value)
            elif self._quant_cfg.policy.has_property(QuantizationProperty.PER_CHANNEL):
                raise TypeError('Isotone Observer is not designed for channelwise quantization.')
            else:
                raise TypeError('Isotone Observer only work with per-tensor or per-channel quantize policy.')

    def render_quantization_config(self):
        # If TQC is not prepared for calibration, just skip this execution.
        if self._quant_cfg.state not in {QuantizationStates.INITIAL}:
            return

        device = self._cache[-1].device
        collected = torch.cat(self._cache, dim=0)
        collected = collected.cpu().numpy()
        s_candidates = []

        for l1, l2 in collected:
            if self._quant_cfg.policy.has_property(QuantizationProperty.SYMMETRICAL):
                l1, l2 = abs(l1), abs(l2)

            scale_min = max(l2 / (self._quant_cfg.quant_max - 0.51), 0)
            scale_max = 2 * (l1 - max(l2, 0))
            if scale_max > scale_min and l1 > 0:
                s_candidates.append((scale_min, 'min'))
                s_candidates.append((scale_max, 'max'))

        if len(s_candidates) <= 0:
            # fall back to min-max calibration
            scale, offset = minmax_to_scale_offset(min_val=0, max_val=l1, config=self._quant_cfg)
            self._quant_cfg.scale = torch.tensor([scale], dtype=torch.float32, device=device).squeeze(0)
            self._quant_cfg.offset = torch.tensor([offset], dtype=torch.float32, device=device).squeeze(0)
            self._quant_cfg.state = QuantizationStates.ACTIVATED
            ppq_warning(
                f'There is no way to make clear classification on {self._watch_on.name} under int8 quantization.\n'
                f'变量 {self._watch_on.name} 无法进行保序校准，在校准数据集上无法保证分类正确性，请检查数据。'
            )
            return

        s_candidates, best_satisfied, satisfied = sorted(s_candidates), 0, 0
        for s_candidate, T in s_candidates:
            if T == 'min':
                satisfied += 1
            if T == 'max':
                satisfied -= 1

            if satisfied > best_satisfied:
                best_satisfied = satisfied
                best_scale = s_candidate

        self._quant_cfg.scale = torch.tensor([best_scale], dtype=torch.float32, device=device).squeeze(0)
        self._quant_cfg.offset = torch.tensor([0], dtype=torch.float32, device=device).squeeze(0)
        self._quant_cfg.state = QuantizationStates.ACTIVATED
        self.s_candidates = s_candidates



from typing import Union

import torch

from esp_ppq.core import (
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.IR import BaseGraph, GraphCommandProcessor, Operation

from .base import BaseQuantizer


class TensorRTQuantizer(BaseQuantizer):
    def __init__(self, graph: Union[BaseGraph, GraphCommandProcessor]) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = -128
        self._quant_max = +127

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        OQC = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
        )

        if operation.type in {'Conv', 'ConvTranspose', 'Gemm', 'MatMul', 'PPQBiasFusedMatMul'}:
            # base_quant_config.output_quantization_config[0].state = QuantizationStates.FP32
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel quant-config.
            assert operation.num_of_input > 0, 'Seems you got a Conv layer with no parameters.'

            # first parameter must exits, for conv layer it will be conv_weight
            # layout: [out_channel, in_channel, kernel_size, kernel_size]
            if operation.type in {'Conv', 'ConvTranspose'}:
                if operation.inputs[1].is_parameter:
                    conv_weight_config = OQC.input_quantization_config[1]
                    conv_weight_config.policy = QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL
                        + QuantizationProperty.LINEAR
                        + QuantizationProperty.PER_CHANNEL
                    )
                    conv_weight_config.channel_axis = 1 if operation.type == 'ConvTranspose' else 0
                    conv_weight_config.observer_algorithm = 'minmax'

            # first parameter must exits, for gemm layer it will be gemm_weight
            # layout: [in_dim, out_dim]
            elif operation.type in {'Gemm', 'MatMul', 'PPQBiasFusedMatMul'}:
                if operation.inputs[1].is_parameter:
                    gemm_weight_config = OQC.input_quantization_config[1]
                    gemm_weight_config.policy = QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL
                        + QuantizationProperty.LINEAR
                        + QuantizationProperty.PER_CHANNEL
                    )
                    gemm_weight_config.channel_axis = 0
                    gemm_weight_config.observer_algorithm = 'minmax'

            if operation.num_of_input > 2:
                bias_config = OQC.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

        if operation.type == 'LayerNormalization':
            # Layernorm - gamma and beta need to be FP32
            for TQC in OQC.input_quantization_config[1:]:
                TQC.state = QuantizationStates.FP32

        if operation.type == 'Attention':
            # Attention - Only input and weight need to be quantized.
            for TQC in OQC.input_quantization_config[2:]:
                TQC.state = QuantizationStates.FP32

        return OQC

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.TRT_INT8

    @property
    def quant_operation_types(self) -> set:
        return {
            'Conv',
            'ConvTranspose',
            'Gemm',
            'Relu',
            'PRelu',
            'Clip',
            'Pad',
            'Resize',
            'MaxPool',
            'AveragePool',
            'GlobalMaxPool',
            'GlobalAveragePool',
            'Softmax',
            'Mul',
            'Add',
            'Max',
            'Sub',
            'Div',
            'Reshape',
            'LeakyRelu',
            'Concat',
            'Sigmoid',
            'Interp',
            'ReduceMean',
            'Transpose',
            'Slice',
            'Flatten',
            'HardSwish',
            'HardSigmoid',
            'MatMul',
            'Attention',
            'LayerNormalization',
            'Gelu',
            'PPQBiasFusedMatMul',
        }

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip', 'Swish', 'SoftPlus', 'Sigmoid', 'Gelu'}


class TensorRTQuantizer_InputOnly(BaseQuantizer):
    """
    This is another version of TensorRT Int8 Quantizer.
        Only Quantize Conv, Gemm, AveragePooling.
        Only Quantize Input Of those layer.
    """

    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = -128
        self._quant_max = +127

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
        )

        if operation.type in {'Conv', 'ConvTranspose', 'Gemm', 'MatMul', 'PPQBiasFusedMatMul'}:
            base_quant_config.output_quantization_config[0].state = QuantizationStates.FP32
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel quant-config.
            assert operation.num_of_input > 0, 'Seems you got a Conv layer with no parameters.'

            # first parameter must exits, for conv layer it will be conv_weight
            # layout: [out_channel, in_channel, kernel_size, kernel_size]
            if operation.type in {'Conv', 'ConvTranspose'}:
                if operation.inputs[1].is_parameter:
                    conv_weight_config = base_quant_config.input_quantization_config[1]
                    conv_weight_config.policy = QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL
                        + QuantizationProperty.LINEAR
                        + QuantizationProperty.PER_CHANNEL
                    )
                    conv_weight_config.channel_axis = 1 if operation.type == 'ConvTranspose' else 0
                    conv_weight_config.observer_algorithm = 'minmax'
            # first parameter must exits, for gemm layer it will be gemm_weight
            # layout: [in_dim, out_dim]
            elif operation.type in {'Gemm', 'MatMul', 'PPQBiasFusedMatMul'}:
                if operation.inputs[1].is_parameter:
                    gemm_weight_config = base_quant_config.input_quantization_config[1]
                    gemm_weight_config.policy = QuantizationPolicy(
                        QuantizationProperty.SYMMETRICAL
                        + QuantizationProperty.LINEAR
                        + QuantizationProperty.PER_CHANNEL
                    )
                    gemm_weight_config.channel_axis = 0
                    gemm_weight_config.observer_algorithm = 'minmax'
            # if operation has bias
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32
        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.TRT_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        return {'Conv', 'Gemm', 'ConvTranspose', 'MatMul', 'AveragePool', 'GlobalAveragePool', 'PPQBiasFusedMatMul'}

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        # 这个量化器只对输入定点，不需要考虑激活函数融合
        return {'Relu', 'Clip'}



import json

from esp_ppq.core import DataType, QuantizationProperty, QuantizationStates, TargetPlatform, TensorQuantizationConfig
from esp_ppq.IR import BaseGraph
from esp_ppq.IR.quantize import QuantableOperation

from .onnx_exporter import OnnxExporter
from .util import convert_value


def convert_type(platform: TargetPlatform) -> str:
    if platform == TargetPlatform.PPL_CUDA_INT8:
        return 'INT8'
    if platform == TargetPlatform.SOI:
        return None
    if platform == TargetPlatform.FP32:
        return None
    raise TypeError(f'Unsupported platform type. ({str(platform)})')


class PPLBackendExporter(OnnxExporter):
    def export_quantization_config(self, config_path: str, graph: BaseGraph):
        var_quant_info_recorder, op_platform_recorder = {}, {}
        for operation in graph.operations.values():
            if not isinstance(operation, QuantableOperation):
                continue
            for config, var in operation.config_with_variable:
                if not config.can_export():
                    continue

                # PATCH 2021.11.25
                # REMOVE BIAS FROM CONFIGURATION
                if config.num_of_bits > 8:
                    continue

                if config.state in {
                    QuantizationStates.FP32,
                }:
                    continue
                # Simply override recorder is acceptable here,
                # we do not support mix precision quantization for CUDA backend now.
                # All configurations for this variable should keep identical towards each other.

                if config.state == QuantizationStates.PASSIVE and var.name in var_quant_info_recorder:
                    continue
                var_quant_info_recorder[var.name] = config

        # ready to render config to json.
        for var in var_quant_info_recorder:
            config = var_quant_info_recorder[var]
            assert isinstance(config, TensorQuantizationConfig)
            tensorwise = config.policy.has_property(QuantizationProperty.PER_TENSOR)
            var_quant_info_recorder[var] = {
                'bit_width': config.num_of_bits,
                'per_channel': config.policy.has_property(QuantizationProperty.PER_CHANNEL),
                'quant_flag': True,
                'sym': config.policy.has_property(QuantizationProperty.SYMMETRICAL),
                'scale': convert_value(config.scale, tensorwise, DataType.FP32),
                'zero_point': convert_value(config.offset, tensorwise, DataType.INT32),
                'tensor_min': convert_value(
                    config.scale * (config.quant_min - config.offset), tensorwise, DataType.FP32
                ),
                'tensor_max': convert_value(
                    config.scale * (config.quant_max - config.offset), tensorwise, DataType.FP32
                ),
                'q_min': config.quant_min,
                'q_max': config.quant_max,
                'hash': config.__hash__(),
                'dominator': config.dominated_by.__hash__(),
            }

        for op in graph.operations.values():
            if convert_type(op.platform) is not None:
                op_platform_recorder[op.name] = {'data_type': convert_type(op.platform)}

        exports = {'quant_info': var_quant_info_recorder, 'op_info': op_platform_recorder}

        with open(file=config_path, mode='w') as file:
            json.dump(exports, file, indent=4)



"""这个脚本将教会你如何使用 PPQ 量化自定义算子"""

import torch

from esp_ppq import *
from esp_ppq.api import *
from esp_ppq.quantization.quantizer import TensorRTQuantizer

B = 1
T = 64
MODEL_PATH = 'models\encoder_ln.onnx'


def generate_samples(num_of_samples: int = 32):
    """生成样本数据，把这个函数改成真实数据读入就可以完成量化了
    这个语音数据量很小 我建议你把整个数据集直接全部送上CUDA
    """
    sample = {'speech': torch.rand(size=[B, T, 80]).float().cuda(), 'speech_lengths': torch.ones(size=[B]).int().cuda()}
    samples = [sample for _ in range(num_of_samples)]
    return samples


SAMPLES = generate_samples()


# 定义一个自己的量化器，定制量化行为，继承于 TensorRTQuantizer 量化器
class MyTensorRTQuantizer(TensorRTQuantizer):
    @property
    def quant_operation_types(self) -> set:
        """覆盖 quant_operation_types  自定义需要量化的算子"""
        return {'LayerNormPlugin'}

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        config = super().init_quantize_config(operation=operation)
        """针对 LayerNormPlugin 生成量化配置信息"""
        if operation.type == 'LayerNormPlugin':
            wconfig = config.input_quantization_config[1]  # weight config
            bconfig = config.input_quantization_config[2]

            wconfig.policy = QuantizationPolicy(  # weight 做 Per channel 量化
                QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
            )
            bconfig.state = QuantizationStates.FP32  # bias 不量化
            '''
            # 将 weight config 升级为 ChannelwiseTensorQuantizationConfig
            config.input_quantization_config[1] = (
                ChannelwiseTensorQuantizationConfig.convert_from_tensor_config(
                    convert_from = wconfig,
                    offsets = None, scales = None,
                    channel_axis = 0))
            '''
            config.input_quantization_config[1].observer_algorithm = 'Minmax'
        return config


def layernorm_forward(op: Operation, values: List[torch.Tensor], ctx=None, **kwargs):
    """自定义算子的前向传播函数"""
    return values[0]


register_operation_handler(
    handler=layernorm_forward, operation_type='LayerNormPlugin', platform=TargetPlatform.TRT_INT8
)

register_network_quantizer(quantizer=MyTensorRTQuantizer, platform=TargetPlatform.TRT_INT8)

with ENABLE_CUDA_KERNEL():
    QS = QuantizationSettingFactory.trt_setting()
    ir = load_onnx_graph(onnx_import_file=MODEL_PATH)
    # 默认调度失效，直接手动调度所有 LayerNormPlugin 送上量化区
    for op in ir.operations.values():
        if op.type == 'LayerNormPlugin':
            QS.dispatching_table.append(operation=op.name, platform=TargetPlatform.TRT_INT8)

    qir = quantize_native_model(
        model=ir,
        calib_dataloader=SAMPLES,
        calib_steps=32,
        input_shape=None,
        inputs=SAMPLES[0],
        platform=TargetPlatform.TRT_INT8,
        setting=QS,
    )

    graphwise_error_analyse(graph=qir, running_device='cuda', dataloader=SAMPLES)

    export_ppq_graph(graph=qir, platform=TargetPlatform.ONNXRUNTIME, graph_save_to='quantized.onnx')



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class TypeInfo(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = TypeInfo()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsTypeInfo(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # TypeInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # TypeInfo
    def ValueType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # TypeInfo
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            from flatbuffers.table import Table
            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None

    # TypeInfo
    def Denotation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

def TypeInfoStart(builder):
    builder.StartObject(3)

def Start(builder):
    TypeInfoStart(builder)

def TypeInfoAddValueType(builder, valueType):
    builder.PrependUint8Slot(0, valueType, 0)

def AddValueType(builder, valueType):
    TypeInfoAddValueType(builder, valueType)

def TypeInfoAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0)

def AddValue(builder, value):
    TypeInfoAddValue(builder, value)

def TypeInfoAddDenotation(builder, denotation):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(denotation), 0)

def AddDenotation(builder, denotation):
    TypeInfoAddDenotation(builder, denotation)

def TypeInfoEnd(builder):
    return builder.EndObject()

def End(builder):
    return TypeInfoEnd(builder)

import FlatBuffers.Dl.MapType
import FlatBuffers.Dl.OptionalType
import FlatBuffers.Dl.SequenceType
import FlatBuffers.Dl.TensorTypeAndShape
import FlatBuffers.Dl.TypeInfoValue
try:
    from typing import Union
except:
    pass

class TypeInfoT(object):

    # TypeInfoT
    def __init__(self):
        self.valueType = 0  # type: int
        self.value = None  # type: Union[None, FlatBuffers.Dl.TensorTypeAndShape.TensorTypeAndShapeT, FlatBuffers.Dl.SequenceType.SequenceTypeT, FlatBuffers.Dl.MapType.MapTypeT, FlatBuffers.Dl.OptionalType.OptionalTypeT]
        self.denotation = None  # type: str

    @classmethod
    def InitFromBuf(cls, buf, pos):
        typeInfo = TypeInfo()
        typeInfo.Init(buf, pos)
        return cls.InitFromObj(typeInfo)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, typeInfo):
        x = TypeInfoT()
        x._UnPack(typeInfo)
        return x

    # TypeInfoT
    def _UnPack(self, typeInfo):
        if typeInfo is None:
            return
        self.valueType = typeInfo.ValueType()
        self.value = FlatBuffers.Dl.TypeInfoValue.TypeInfoValueCreator(self.valueType, typeInfo.Value())
        self.denotation = typeInfo.Denotation()

    # TypeInfoT
    def Pack(self, builder):
        if self.value is not None:
            value = self.value.Pack(builder)
        if self.denotation is not None:
            denotation = builder.CreateString(self.denotation)
        TypeInfoStart(builder)
        TypeInfoAddValueType(builder, self.valueType)
        if self.value is not None:
            TypeInfoAddValue(builder, value)
        if self.denotation is not None:
            TypeInfoAddDenotation(builder, denotation)
        typeInfo = TypeInfoEnd(builder)
        return typeInfo



from typing import Union

import torch

from esp_ppq.core import (
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.IR import BaseGraph, Operation

from .base import BaseQuantizer


class RKNN_PerTensorQuantizer(BaseQuantizer):
    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = 0
        self._quant_max = int(pow(2, self._num_of_bits) - 1)

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
        )

        if operation.type in {'Conv', 'Gemm'}:
            # if operation has bias
            if operation.num_of_input == 3:
                bias_config = base_quant_config.input_quantization_config[-1]
                # bias should be quantized with 32 bits
                # in python3, int indicates long long in C++
                # so that it has enough precision to represent a number like 2^32
                # however, it may cause a scale underflow
                # here we give bias a 30 bits precision, which is pettery enough in all cases
                bias_config.num_of_bits = 32
                bias_config.quant_max = int(pow(2, 30 - 1))
                bias_config.quant_min = -int(pow(2, 30 - 1))
                bias_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
                )
                bias_config.state = QuantizationStates.PASSIVE_INIT
            for config in base_quant_config.input_quantization_config[1:]:
                config.observer_algorithm = 'minmax'

        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            base_quant_config.is_active_quant_op = False
        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.RKNN_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        ONNX_OOS_OPSET = {
            'Conv',
            'GlobalAveragePool',
            'AveragePool',
            'Relu',
            'Add',
            'Mul',
            'Clip',
            'Sigmoid',
            'MatMul',
            'Gemm',
            'Concat',
            'LeakyRelu',
        }
        ONNX_OOS_OPSET.update(PASSIVE_OPERATIONS)
        return ONNX_OOS_OPSET

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.ASYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip'}


class RKNN_PerChannelQuantizer(BaseQuantizer):
    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = 0
        self._quant_max = int(pow(2, self._num_of_bits) - 1)

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
        )

        if operation.type in {'Conv', 'MatMul'}:
            # set all parameters within Conv, ConvTranspose, Gemm to per-channel quant-config.
            assert operation.num_of_input > 0, 'Seems you got a Conv layer with no parameters.'

            # first parameter must exits, for conv layer it will be conv_weight
            # layout: [out_channel, in_channel, kernel_size, kernel_size]
            if operation.type in {'Conv'}:
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.quant_max = +127
                conv_weight_config.quant_min = -127
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                conv_weight_config.channel_axis = 0
                conv_weight_config.observer_algorithm = 'minmax'
            # first parameter must exits, for gemm layer it will be gemm_weight
            # layout: [in_dim, out_dim]
            else:
                for index, input_var in enumerate(operation.inputs):
                    if input_var.is_parameter:
                        matmul_weight_config = base_quant_config.input_quantization_config[index]
                        matmul_weight_config.quant_max = +127
                        matmul_weight_config.quant_min = -127
                        matmul_weight_config.policy = QuantizationPolicy(
                            QuantizationProperty.SYMMETRICAL
                            + QuantizationProperty.LINEAR
                            + QuantizationProperty.PER_CHANNEL
                        )
                        matmul_weight_config.channel_axis = 1
                        matmul_weight_config.observer_algorithm = 'minmax'
            # if operation has bias
            if operation.type in {'Conv', 'Gemm'} and operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                bias_config.num_of_bits = 30
                bias_config.quant_max = int(pow(2, 30 - 1))
                bias_config.quant_min = -int(pow(2, 30 - 1))
                bias_config.state = QuantizationStates.PASSIVE_INIT
                bias_config.channel_axis = 0
                bias_config.observer_algorithm = 'minmax'

        if operation.type in PASSIVE_OPERATIONS:
            base_quant_config.is_active_quant_op = False

        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.RKNN_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        ONNX_OOS_OPSET = {
            'Conv',
            'GlobalAveragePool',
            'AveragePool',
            'Relu',
            'Add',
            'Mul',
            'Clip',
            'Sigmoid',
            'MatMul',
            'Gemm',
            'Concat',
            'LeakyRelu',
        }
        ONNX_OOS_OPSET.update(PASSIVE_OPERATIONS)
        return ONNX_OOS_OPSET

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.ASYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        return {'Relu', 'Clip'}



# ---------------------------------------------------------------
# 这个脚本向你展示了如何使用绕过那些跟量化没什么关系的算子
#   当你的算子处于网络的最后，其后也没有什么需要量化的算子了
#   你就可以给它定义一个假的 forward 函数，从而帮助 PPQ 完成量化
#   PPQ 不再需要收集其后的数据信息，所以错误额计算过程也能得到正确的量化结果
#
#   当然如果你的自定义算子如果会干涉到量化过程，你还是需要向 PPQ 提供一个的执行函数
# ---------------------------------------------------------------

# For this inference test, all test data is randomly picked.
# If you want to use real data, just rewrite the defination of SAMPLES
import torch
from torch.utils.data import DataLoader

from esp_ppq import *
from esp_ppq.api import *

BATCHSIZE = 32
INPUT_SHAPE = [3, 224, 224]
DEVICE = 'cuda'  # only cuda is fully tested :(  For other executing device there might be bugs.
PLATFORM = TargetPlatform.TRT_INT8  # identify a target platform for your network.


def load_calibration_dataset() -> Iterable:
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


def happy_forward_func(op: Operation, values, ctx, **kwargs):
    """你必须保证函数签名满足要求，即函数的输入和返回值都满足 PPQ 的系统要求

    你的执行函数将接收 op, values, ctx 三个元素作为输入
    其中 op 反应了当前执行的算子信息，values是一个数组，包含了算子所有输入
    ctx 是 PPQ 执行上下文

    你将返回一个 torch.Tensor 或者多个 torch.Tensor 作为结果
    这取决于你的算子在onnx中有多少个输出
    """
    return torch.zeros(size=[1, 100, 5]), torch.zeros(size=[1, 100])


# ---------------------------------------------
# 注册一个假的函数让我们绕过 nms
# ---------------------------------------------
register_operation_handler(happy_forward_func, operation_type="TRTBatchedNMS", platform=TargetPlatform.FP32)

quant_setting = QuantizationSettingFactory.default_setting()

# For pytorch user, just dump your network to disk with onnx first
unquantized = load_onnx_graph(onnx_import_file='Output/onnx.model')

# Load training data for creating a calibration dataloader.
calibration_dataset = load_calibration_dataset()
calibration_dataloader = DataLoader(dataset=calibration_dataset, batch_size=BATCHSIZE, shuffle=False)

# quantize your model.
quantized = quantize_native_model(
    model=unquantized,
    calib_dataloader=calibration_dataloader,
    calib_steps=32,
    input_shape=[BATCHSIZE] + INPUT_SHAPE,
    setting=quant_setting,
    collate_fn=collate_fn,
    platform=PLATFORM,
    device=DEVICE,
    verbose=0,
)

# Quantization Result is a PPQ BaseGraph instance.
assert isinstance(quantized, BaseGraph)

# export quantized graph.
export_ppq_graph(
    graph=quantized,
    platform=PLATFORM,
    graph_save_to='Output/quantized(onnx).onnx',
    config_save_to='Output/quantized(onnx).json',
)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class Attribute(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Attribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsAttribute(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    # Attribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Attribute
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Attribute
    def RefAttrName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Attribute
    def DocString(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Attribute
    def AttrType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Attribute
    def F(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from FlatBuffers.Dl.AttributeF import AttributeF
            obj = AttributeF()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Attribute
    def I(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = o + self._tab.Pos
            from FlatBuffers.Dl.AttributeI import AttributeI
            obj = AttributeI()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Attribute
    def S(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # Attribute
    def SAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Attribute
    def SLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def SIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0

    # Attribute
    def T(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from FlatBuffers.Dl.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Attribute
    def G(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from FlatBuffers.Dl.Graph import Graph
            obj = Graph()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Attribute
    def Tp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from FlatBuffers.Dl.TypeInfo import TypeInfo
            obj = TypeInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Attribute
    def Floats(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Float32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Attribute
    def FloatsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float32Flags, o)
        return 0

    # Attribute
    def FloatsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def FloatsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        return o == 0

    # Attribute
    def Ints(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Int64Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8))
        return 0

    # Attribute
    def IntsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Int64Flags, o)
        return 0

    # Attribute
    def IntsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def IntsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        return o == 0

    # Attribute
    def Strings(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # Attribute
    def StringsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def StringsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        return o == 0

    # Attribute
    def Tensors(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Tensor import Tensor
            obj = Tensor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Attribute
    def TensorsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def TensorsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        return o == 0

    # Attribute
    def Graphs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.Graph import Graph
            obj = Graph()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Attribute
    def GraphsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def GraphsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(32))
        return o == 0

    # Attribute
    def TypeProtos(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from FlatBuffers.Dl.TypeInfo import TypeInfo
            obj = TypeInfo()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Attribute
    def TypeProtosLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def TypeProtosIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(34))
        return o == 0

def AttributeStart(builder):
    builder.StartObject(16)

def Start(builder):
    AttributeStart(builder)

def AttributeAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)

def AddName(builder, name):
    AttributeAddName(builder, name)

def AttributeAddRefAttrName(builder, refAttrName):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(refAttrName), 0)

def AddRefAttrName(builder, refAttrName):
    AttributeAddRefAttrName(builder, refAttrName)

def AttributeAddDocString(builder, docString):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(docString), 0)

def AddDocString(builder, docString):
    AttributeAddDocString(builder, docString)

def AttributeAddAttrType(builder, attrType):
    builder.PrependInt32Slot(3, attrType, 0)

def AddAttrType(builder, attrType):
    AttributeAddAttrType(builder, attrType)

def AttributeAddF(builder, f):
    builder.PrependStructSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(f), 0)

def AddF(builder, f):
    AttributeAddF(builder, f)

def AttributeAddI(builder, i):
    builder.PrependStructSlot(5, flatbuffers.number_types.UOffsetTFlags.py_type(i), 0)

def AddI(builder, i):
    AttributeAddI(builder, i)

def AttributeAddS(builder, s):
    builder.PrependUOffsetTRelativeSlot(6, flatbuffers.number_types.UOffsetTFlags.py_type(s), 0)

def AddS(builder, s):
    AttributeAddS(builder, s)

def AttributeStartSVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)

def StartSVector(builder, numElems):
    return AttributeStartSVector(builder, numElems)

def AttributeAddT(builder, t):
    builder.PrependUOffsetTRelativeSlot(7, flatbuffers.number_types.UOffsetTFlags.py_type(t), 0)

def AddT(builder, t):
    AttributeAddT(builder, t)

def AttributeAddG(builder, g):
    builder.PrependUOffsetTRelativeSlot(8, flatbuffers.number_types.UOffsetTFlags.py_type(g), 0)

def AddG(builder, g):
    AttributeAddG(builder, g)

def AttributeAddTp(builder, tp):
    builder.PrependUOffsetTRelativeSlot(9, flatbuffers.number_types.UOffsetTFlags.py_type(tp), 0)

def AddTp(builder, tp):
    AttributeAddTp(builder, tp)

def AttributeAddFloats(builder, floats):
    builder.PrependUOffsetTRelativeSlot(10, flatbuffers.number_types.UOffsetTFlags.py_type(floats), 0)

def AddFloats(builder, floats):
    AttributeAddFloats(builder, floats)

def AttributeStartFloatsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartFloatsVector(builder, numElems):
    return AttributeStartFloatsVector(builder, numElems)

def AttributeAddInts(builder, ints):
    builder.PrependUOffsetTRelativeSlot(11, flatbuffers.number_types.UOffsetTFlags.py_type(ints), 0)

def AddInts(builder, ints):
    AttributeAddInts(builder, ints)

def AttributeStartIntsVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)

def StartIntsVector(builder, numElems):
    return AttributeStartIntsVector(builder, numElems)

def AttributeAddStrings(builder, strings):
    builder.PrependUOffsetTRelativeSlot(12, flatbuffers.number_types.UOffsetTFlags.py_type(strings), 0)

def AddStrings(builder, strings):
    AttributeAddStrings(builder, strings)

def AttributeStartStringsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartStringsVector(builder, numElems):
    return AttributeStartStringsVector(builder, numElems)

def AttributeAddTensors(builder, tensors):
    builder.PrependUOffsetTRelativeSlot(13, flatbuffers.number_types.UOffsetTFlags.py_type(tensors), 0)

def AddTensors(builder, tensors):
    AttributeAddTensors(builder, tensors)

def AttributeStartTensorsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartTensorsVector(builder, numElems):
    return AttributeStartTensorsVector(builder, numElems)

def AttributeAddGraphs(builder, graphs):
    builder.PrependUOffsetTRelativeSlot(14, flatbuffers.number_types.UOffsetTFlags.py_type(graphs), 0)

def AddGraphs(builder, graphs):
    AttributeAddGraphs(builder, graphs)

def AttributeStartGraphsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartGraphsVector(builder, numElems):
    return AttributeStartGraphsVector(builder, numElems)

def AttributeAddTypeProtos(builder, typeProtos):
    builder.PrependUOffsetTRelativeSlot(15, flatbuffers.number_types.UOffsetTFlags.py_type(typeProtos), 0)

def AddTypeProtos(builder, typeProtos):
    AttributeAddTypeProtos(builder, typeProtos)

def AttributeStartTypeProtosVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartTypeProtosVector(builder, numElems):
    return AttributeStartTypeProtosVector(builder, numElems)

def AttributeEnd(builder):
    return builder.EndObject()

def End(builder):
    return AttributeEnd(builder)

import FlatBuffers.Dl.AttributeF
import FlatBuffers.Dl.AttributeI
import FlatBuffers.Dl.Graph
import FlatBuffers.Dl.Tensor
import FlatBuffers.Dl.TypeInfo
try:
    from typing import List, Optional
except:
    pass

class AttributeT(object):

    # AttributeT
    def __init__(self):
        self.name = None  # type: str
        self.refAttrName = None  # type: str
        self.docString = None  # type: str
        self.attrType = 0  # type: int
        self.f = None  # type: Optional[FlatBuffers.Dl.AttributeF.AttributeFT]
        self.i = None  # type: Optional[FlatBuffers.Dl.AttributeI.AttributeIT]
        self.s = None  # type: List[int]
        self.t = None  # type: Optional[FlatBuffers.Dl.Tensor.TensorT]
        self.g = None  # type: Optional[FlatBuffers.Dl.Graph.GraphT]
        self.tp = None  # type: Optional[FlatBuffers.Dl.TypeInfo.TypeInfoT]
        self.floats = None  # type: List[float]
        self.ints = None  # type: List[int]
        self.strings = None  # type: List[str]
        self.tensors = None  # type: List[FlatBuffers.Dl.Tensor.TensorT]
        self.graphs = None  # type: List[FlatBuffers.Dl.Graph.GraphT]
        self.typeProtos = None  # type: List[FlatBuffers.Dl.TypeInfo.TypeInfoT]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        attribute = Attribute()
        attribute.Init(buf, pos)
        return cls.InitFromObj(attribute)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, attribute):
        x = AttributeT()
        x._UnPack(attribute)
        return x

    # AttributeT
    def _UnPack(self, attribute):
        if attribute is None:
            return
        self.name = attribute.Name()
        self.refAttrName = attribute.RefAttrName()
        self.docString = attribute.DocString()
        self.attrType = attribute.AttrType()
        if attribute.F() is not None:
            self.f = FlatBuffers.Dl.AttributeF.AttributeFT.InitFromObj(attribute.F())
        if attribute.I() is not None:
            self.i = FlatBuffers.Dl.AttributeI.AttributeIT.InitFromObj(attribute.I())
        if not attribute.SIsNone():
            if np is None:
                self.s = []
                for i in range(attribute.SLength()):
                    self.s.append(attribute.S(i))
            else:
                self.s = attribute.SAsNumpy()
        if attribute.T() is not None:
            self.t = FlatBuffers.Dl.Tensor.TensorT.InitFromObj(attribute.T())
        if attribute.G() is not None:
            self.g = FlatBuffers.Dl.Graph.GraphT.InitFromObj(attribute.G())
        if attribute.Tp() is not None:
            self.tp = FlatBuffers.Dl.TypeInfo.TypeInfoT.InitFromObj(attribute.Tp())
        if not attribute.FloatsIsNone():
            if np is None:
                self.floats = []
                for i in range(attribute.FloatsLength()):
                    self.floats.append(attribute.Floats(i))
            else:
                self.floats = attribute.FloatsAsNumpy()
        if not attribute.IntsIsNone():
            if np is None:
                self.ints = []
                for i in range(attribute.IntsLength()):
                    self.ints.append(attribute.Ints(i))
            else:
                self.ints = attribute.IntsAsNumpy()
        if not attribute.StringsIsNone():
            self.strings = []
            for i in range(attribute.StringsLength()):
                self.strings.append(attribute.Strings(i))
        if not attribute.TensorsIsNone():
            self.tensors = []
            for i in range(attribute.TensorsLength()):
                if attribute.Tensors(i) is None:
                    self.tensors.append(None)
                else:
                    tensor_ = FlatBuffers.Dl.Tensor.TensorT.InitFromObj(attribute.Tensors(i))
                    self.tensors.append(tensor_)
        if not attribute.GraphsIsNone():
            self.graphs = []
            for i in range(attribute.GraphsLength()):
                if attribute.Graphs(i) is None:
                    self.graphs.append(None)
                else:
                    graph_ = FlatBuffers.Dl.Graph.GraphT.InitFromObj(attribute.Graphs(i))
                    self.graphs.append(graph_)
        if not attribute.TypeProtosIsNone():
            self.typeProtos = []
            for i in range(attribute.TypeProtosLength()):
                if attribute.TypeProtos(i) is None:
                    self.typeProtos.append(None)
                else:
                    typeInfo_ = FlatBuffers.Dl.TypeInfo.TypeInfoT.InitFromObj(attribute.TypeProtos(i))
                    self.typeProtos.append(typeInfo_)

    # AttributeT
    def Pack(self, builder):
        if self.name is not None:
            name = builder.CreateString(self.name)
        if self.refAttrName is not None:
            refAttrName = builder.CreateString(self.refAttrName)
        if self.docString is not None:
            docString = builder.CreateString(self.docString)
        if self.s is not None:
            if np is not None and type(self.s) is np.ndarray:
                s = builder.CreateNumpyVector(self.s)
            else:
                AttributeStartSVector(builder, len(self.s))
                for i in reversed(range(len(self.s))):
                    builder.PrependUint8(self.s[i])
                s = builder.EndVector()
        if self.t is not None:
            t = self.t.Pack(builder)
        if self.g is not None:
            g = self.g.Pack(builder)
        if self.tp is not None:
            tp = self.tp.Pack(builder)
        if self.floats is not None:
            if np is not None and type(self.floats) is np.ndarray:
                floats = builder.CreateNumpyVector(self.floats)
            else:
                AttributeStartFloatsVector(builder, len(self.floats))
                for i in reversed(range(len(self.floats))):
                    builder.PrependFloat32(self.floats[i])
                floats = builder.EndVector()
        if self.ints is not None:
            if np is not None and type(self.ints) is np.ndarray:
                ints = builder.CreateNumpyVector(self.ints)
            else:
                AttributeStartIntsVector(builder, len(self.ints))
                for i in reversed(range(len(self.ints))):
                    builder.PrependInt64(self.ints[i])
                ints = builder.EndVector()
        if self.strings is not None:
            stringslist = []
            for i in range(len(self.strings)):
                stringslist.append(builder.CreateString(self.strings[i]))
            AttributeStartStringsVector(builder, len(self.strings))
            for i in reversed(range(len(self.strings))):
                builder.PrependUOffsetTRelative(stringslist[i])
            strings = builder.EndVector()
        if self.tensors is not None:
            tensorslist = []
            for i in range(len(self.tensors)):
                tensorslist.append(self.tensors[i].Pack(builder))
            AttributeStartTensorsVector(builder, len(self.tensors))
            for i in reversed(range(len(self.tensors))):
                builder.PrependUOffsetTRelative(tensorslist[i])
            tensors = builder.EndVector()
        if self.graphs is not None:
            graphslist = []
            for i in range(len(self.graphs)):
                graphslist.append(self.graphs[i].Pack(builder))
            AttributeStartGraphsVector(builder, len(self.graphs))
            for i in reversed(range(len(self.graphs))):
                builder.PrependUOffsetTRelative(graphslist[i])
            graphs = builder.EndVector()
        if self.typeProtos is not None:
            typeProtoslist = []
            for i in range(len(self.typeProtos)):
                typeProtoslist.append(self.typeProtos[i].Pack(builder))
            AttributeStartTypeProtosVector(builder, len(self.typeProtos))
            for i in reversed(range(len(self.typeProtos))):
                builder.PrependUOffsetTRelative(typeProtoslist[i])
            typeProtos = builder.EndVector()
        AttributeStart(builder)
        if self.name is not None:
            AttributeAddName(builder, name)
        if self.refAttrName is not None:
            AttributeAddRefAttrName(builder, refAttrName)
        if self.docString is not None:
            AttributeAddDocString(builder, docString)
        AttributeAddAttrType(builder, self.attrType)
        if self.f is not None:
            f = self.f.Pack(builder)
            AttributeAddF(builder, f)
        if self.i is not None:
            i = self.i.Pack(builder)
            AttributeAddI(builder, i)
        if self.s is not None:
            AttributeAddS(builder, s)
        if self.t is not None:
            AttributeAddT(builder, t)
        if self.g is not None:
            AttributeAddG(builder, g)
        if self.tp is not None:
            AttributeAddTp(builder, tp)
        if self.floats is not None:
            AttributeAddFloats(builder, floats)
        if self.ints is not None:
            AttributeAddInts(builder, ints)
        if self.strings is not None:
            AttributeAddStrings(builder, strings)
        if self.tensors is not None:
            AttributeAddTensors(builder, tensors)
        if self.graphs is not None:
            AttributeAddGraphs(builder, graphs)
        if self.typeProtos is not None:
            AttributeAddTypeProtos(builder, typeProtos)
        attribute = AttributeEnd(builder)
        return attribute



import numpy as np
import torch

import esp_ppq.lib as PFL
from esp_ppq import TargetPlatform, TorchExecutor, graphwise_error_analyse
from esp_ppq.api import ENABLE_CUDA_KERNEL, export_ppq_graph, load_onnx_graph
from esp_ppq.quantization.optim import *

calibration_dataloader = [torch.rand([1, 3, 640, 640]) for _ in range(32)]

with ENABLE_CUDA_KERNEL():
    graph = load_onnx_graph('Models/yolov5s.v5.onnx')

    quantizer = PFL.Quantizer(platform=TargetPlatform.OPENVINO_INT8, graph=graph)  # 取得 OPENVINO_INT8 所对应的量化器
    dispatching = PFL.Dispatcher(graph=graph).dispatch(  # 生成调度表
        quant_types=quantizer.quant_operation_types
    )

    # ------------------------------------------------------------
    # Yolo5 前面可能有一坨 Slice, Concat 算子
    # 后面可能有一坨后处理算子，我们不希望量化它们，你可用下面的方法将它们解除量化
    # 不同模型中层的名字可能不一样，你需要按照你的模型对它们进行手动修改
    # ------------------------------------------------------------

    # Concat_40 往前的所有算子不量化
    from esp_ppq.IR import SearchableGraph

    search_engine = SearchableGraph(graph)
    for op in search_engine.opset_matching(
        sp_expr=lambda x: x.name == 'Concat_40', rp_expr=lambda x, y: True, ep_expr=None, direction='up'
    ):
        dispatching[op.name] = TargetPlatform.FP32

    # Sigmoid_280 往后的所有算子不量化
    # Sigmoid_246 往后的所有算子不量化
    # Sigmoid_314 往后的所有算子不量化
    for op in search_engine.opset_matching(
        sp_expr=lambda x: x.name in {'Sigmoid_246', 'Sigmoid_280', 'Sigmoid_314'},
        rp_expr=lambda x, y: True,
        ep_expr=None,
        direction='down',
    ):
        dispatching[op.name] = TargetPlatform.FP32

    # 为算子初始化量化信息
    for op in graph.operations.values():
        quantizer.quantize_operation(op_name=op.name, platform=dispatching[op.name])

    # 初始化执行器
    collate_fn = lambda x: x.to('cuda')
    executor = TorchExecutor(graph=graph, device='cuda')
    executor.tracing_operation_meta(inputs=torch.zeros(size=[1, 3, 640, 640]).cuda())
    executor.load_graph(graph=graph)

    # ------------------------------------------------------------
    # 创建优化管线，由于后续还要继续训练我们的模型，我们不能在此处调用
    # ParameterBakingPass()，一旦模型权重完成烘焙，则它们不能被进一步调整
    # ------------------------------------------------------------
    pipeline = PFL.Pipeline(
        [
            QuantizeSimplifyPass(),
            QuantizeFusionPass(activation_type=quantizer.activation_fusion_types),
            ParameterQuantizePass(),
            RuntimeCalibrationPass(),
            PassiveParameterQuantizePass(),
            QuantAlignmentPass(force_overlap=True),
            ParameterBakingPass(),
        ]
    )

with ENABLE_CUDA_KERNEL():
    # 调用管线完成量化
    pipeline.optimize(
        graph=graph,
        dataloader=calibration_dataloader,
        verbose=True,
        calib_steps=32,
        collate_fn=collate_fn,
        executor=executor,
    )

    graphwise_error_analyse(
        graph=graph, running_device='cuda', dataloader=calibration_dataloader, collate_fn=lambda x: x.cuda()
    )

    export_ppq_graph(
        graph=graph,
        platform=TargetPlatform.TRT_INT8,
        graph_save_to='Output/quantized.onnx',
        config_save_to='Output/quantized.json',
    )

from esp_ppq.utils.TensorRTUtil import Benchmark, Profiling, build_engine

build_engine(
    onnx_file='Output/quantized.onnx',
    engine_file='Output/quantized.engine',
    int8=True,
    int8_scale_file='Output/quantized.json',
)
Benchmark('Output/quantized.engine')
Profiling('Output/quantized.engine')



# ---------------------------------------------
# This is a quantizer template for custimizing.
# ---------------------------------------------

from typing import Union

import torch

from esp_ppq.core import (
    PASSIVE_OPERATIONS,
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.IR import BaseGraph, Operation

from .base import BaseQuantizer


class ExtQuantizer(BaseQuantizer):
    """ExtQuantizer 是一个空的量化器模板，你可以在这里面实现你自定义的平台量化逻辑 你需要实现该类所有成员函数与方法，并且使用
    TargetPlatform.EXTENSION 调用这个量化器.

        --> quantize_torch_model(..., platform=EXTENSION)

    ExtQuantizer is a empty quantizer template for you to implement custimized quantization logic.
        Implement all member function and properties, and invoke this with TargetPlatform.EXTENSION

        --> quantize_torch_model(..., platform=EXTENSION)

    Args:
        BaseQuantizer ([type]): [description]
    """

    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)  # do not forget to initialize super class.

        # Quantization basic setting -------------------------
        # update following properties:
        self._num_of_bits = 8  # platform bit width.
        self._quant_min = -128  # int value min
        self._quant_max = +127  # int value max
        # ----------------------------------------------------

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        """使用这个函数来初始化算子的量化配置，该函数会被父类作为接口调用。

        我们提供了一个方便的函数 create_default_quant_config 来帮助你实现相关逻辑
            调用这个函数将直接产生一个默认的量化配置，你可以在此基础上进行修改。

        注意并不是所有算子都可以使用默认量化配置完成量化，有些算子明显具有更加复杂的量化逻辑
            对于这种情况，你需要在此函数中手动创建它们的量化配置信息。

        Initial your quantization configuration for each operation.
            (This function will be invoked by BaseQuantizer as an interface.)

        We provide a helper function called self.create_default_quant_config()
            use this function to create a default quantization configuration.

        However, some operations' quantization logic might be much more complex so that
            default configuration won't meet their requirements. In this situation a
            manually initialized configuration is demended.

        Args:
            operation (Operation): [description]

        Returns:
            OperationQuantizationConfig: [description]
        """

        # create a basic quantization configuration.
        config = self.create_default_quant_config(
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
        )

        # initialize configuration for conv manually
        if operation.type == 'Conv':
            # override initialized config

            if operation.num_of_input == 3:
                bias_config = config.input_quantization_config[-1]
                # bias should be quantized with 32 bits
                # in python3, int indicates long long in C++
                # so that it has enough precision to represent a number like 2^32
                # however, it may cause a scale underflow
                # here we give bias a 30 bits precision, which is pettery enough in all cases
                bias_config.num_of_bits = 30
                bias_config.quant_max = int(pow(2, 30 - 1) - 1)
                bias_config.quant_min = -int(pow(2, 30 - 1))
                bias_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
                )
                bias_config.state = QuantizationStates.PASSIVE_INIT
            for tensor_config in config.input_quantization_config[1:]:
                tensor_config.observer_algorithm = 'minmax'

        # mark some operation as passive op.
        # all quantizations of a passive op will be overlapped during graph merge pass.
        if operation.type in PASSIVE_OPERATIONS:
            # Those op are not active op.
            config.is_active_quant_op = False
        return config

    @property
    def target_platform(self) -> TargetPlatform:
        """target_platform 属性是提供给子图切分使用的， 所有量化区的算子将被调度到这个设备上。

        Property target_platform is acquired by graph dispather.     It states
        where your quantized operation depoly.
        """
        return TargetPlatform.EXTENSION

    @property
    def quant_operation_types(self) -> set:
        """quant_operation_types 指明了所有可以被量化的算子类型。

        聪明的你可能想问了，如果我不是按类型进行量化，而是想特定去量化某几个算子怎么办。
            我建议你通过手写一个调度表来实现这样的功能，只需要把其他算子调度到FP32上就行了。

        请注意，并不是你写在这里的类型就一定会被量化，在量化之前还有调度器进行子图切分，
            只有量化子图上的算子才可以被量化。

        Property quant_operation_types contains all quantable operations' type.

        Notice that PPQ has a dispatching logic before quantizer,
            so to say quantizer is only in charge of quantable subgraph,
            those operation within non-quantable subgraph will never be
            quantized even their type is listed here.
        """
        return {
            'Conv',
            'ConvTranspose',
            'Gemm',
            'Relu',
            'PRelu',
            'Clip',
            'Pad',
            'Resize',
            'MaxPool',
            'AveragePool',
            'GlobalMaxPool',
            'GlobalAveragePool',
            'Mul',
            'Add',
            'Max',
            'Sub',
            'Div',
            'LeakyRelu',
            'Concat',
            'Sigmoid',
            'Slice',
        }

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        """quantize_policy 指明了默认量化策略 被函数 create_default_quant_config 所使用.

        quantize_policy is used by create_default_quant_config(),
            to generate a default quantization configuration.

        Returns:
            QuantizationPolicy: [description]
        """
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        """rounding_policy 指明了默认取整策略 被函数 create_default_quant_config 所使用.

        rounding_policy is used by create_default_quant_config(),
            to generate a default quantization configuration.

        Returns:
            QuantizationPolicy: [description]

        Returns:
            RoundingPolicy: [description]
        """
        return RoundingPolicy.ROUND_HALF_EVEN

    @property
    def activation_fusion_types(self) -> set:
        """activation_fusion_types 指明了所有要参与图融合的激活函数类型 被后续 activation fustion pass 所使用的
        所有被列举在此的激活函数将会尝试和之前的计算节点进行联合定点。

        Returns:
            set: _description_
        """
        return {'Relu', 'Clip'}



import openvino.runtime
import torch
from tqdm import tqdm

from esp_ppq import *

MODEL = 'models\\resnet18.onnx'
INPUT_SHAPE = [1, 3, 224, 224]
SAMPLES = [torch.rand(size=INPUT_SHAPE) for _ in range(256)]  # rewirte this to use real data.

# -------------------------------------------------------------------
# 启动 openvino 进行推理
# -------------------------------------------------------------------
openvino_executor = openvino.runtime.Core()
openvino_results = []
model = openvino_executor.compile_model(model=openvino_executor.read_model(MODEL), device_name="CPU")
for sample in tqdm(SAMPLES, desc='OPENVINO GENERATEING OUTPUTS'):
    openvino_results.append(model([convert_any_to_numpy(sample)]))



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: Dl

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

class AlignedBytes(object):
    __slots__ = ['_tab']

    @classmethod
    def SizeOf(cls):
        return 16

    # AlignedBytes
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # AlignedBytes
    def Bytes(self, j = None):
        if j is None:
            return [self._tab.Get(flatbuffers.number_types.Uint8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0 + i * 1)) for i in range(self.BytesLength())]
        elif j >= 0 and j < self.BytesLength():
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0 + j * 1))
        else:
            return None

    # AlignedBytes
    def BytesAsNumpy(self):
        return self._tab.GetArrayAsNumpy(flatbuffers.number_types.Uint8Flags, self._tab.Pos + 0, self.BytesLength())

    # AlignedBytes
    def BytesLength(self):
        return 16

    # AlignedBytes
    def BytesIsNone(self):
        return False


def CreateAlignedBytes(builder, bytes):
    builder.Prep(16, 16)
    for _idx0 in range(16 , 0, -1):
        builder.PrependUint8(bytes[_idx0-1])
    return builder.Offset()

try:
    from typing import List
except:
    pass

class AlignedBytesT(object):

    # AlignedBytesT
    def __init__(self):
        self.bytes = None  # type: List[int]

    @classmethod
    def InitFromBuf(cls, buf, pos):
        alignedBytes = AlignedBytes()
        alignedBytes.Init(buf, pos)
        return cls.InitFromObj(alignedBytes)

    @classmethod
    def InitFromPackedBuf(cls, buf, pos=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, pos)
        return cls.InitFromBuf(buf, pos+n)

    @classmethod
    def InitFromObj(cls, alignedBytes):
        x = AlignedBytesT()
        x._UnPack(alignedBytes)
        return x

    # AlignedBytesT
    def _UnPack(self, alignedBytes):
        if alignedBytes is None:
            return
        if not alignedBytes.BytesIsNone():
            if np is None:
                self.bytes = []
                for i in range(alignedBytes.BytesLength()):
                    self.bytes.append(alignedBytes.Bytes(i))
            else:
                self.bytes = alignedBytes.BytesAsNumpy()

    # AlignedBytesT
    def Pack(self, builder):
        return CreateAlignedBytes(builder, self.bytes)



from typing import Callable, List

import torch

from esp_ppq.core import (
    NetworkFramework,
    QuantizationPolicy,
    QuantizationProperty,
    RoundingPolicy,
    TargetPlatform,
    TensorQuantizationConfig,
)
from esp_ppq.executor.torch import OPERATION_FORWARD_TABLE
from esp_ppq.IR import BaseGraph, GraphExporter, Variable
from esp_ppq.quantization.observer import BaseTensorObserver, TensorObserverFactroy
from esp_ppq.quantization.optim import QuantizationOptimizationPass, QuantizationOptimizationPipeline
from esp_ppq.quantization.qfunction import PPQuantFunction as QuantFunction
from esp_ppq.quantization.quantizer import BaseQuantizer
from esp_ppq.scheduler import DISPATCHER_TABLE, GraphDispatcher

from .common import __EXPORTERS__, __PARSERS__, __QUANTIZER_COLLECTION__


def Quantizer(platform: TargetPlatform, graph: BaseGraph) -> BaseQuantizer:
    """
    Get a pre-defined Quantizer corresponding to your platform.
    Quantizer in PPQ initializes Tensor Quantization Config for each Operation,
        - it describes how operations are going to be quantized.

    根据目标平台获取一个系统预定义的量化器。

    ## 量化器
    在 PPQ 中，量化器是一个用于为算子初始化量化信息 Tensor Quantization Config 的对象
        - 量化器决定了你的算子是如何被量化的，你也可以设计新的量化器来适配不同的后端推理框架

    在 PPQ 中我们为不同的推理后端设计好了一些预定义的量化器，你可以通过 esp_ppq.lib.Quantizer 来访问它们
    """
    if platform not in __QUANTIZER_COLLECTION__:
        raise KeyError(f'Target Platform {platform} has no related quantizer for now.')
    return __QUANTIZER_COLLECTION__[platform](graph)


def Pipeline(optims: List[QuantizationOptimizationPass]) -> QuantizationOptimizationPipeline:
    """
    Build a Pipeline with given Optimization Passes Collection

    使用给定的量化过程集合创建量化管线。
    """
    return QuantizationOptimizationPipeline(optims)


def Observer(quant_config: TensorQuantizationConfig, variable: Variable = None) -> BaseTensorObserver:
    """
    Get a Calibration Observer based on quant_config.observer_algorithm attribute.

    根据 TQC 中 observer_algorithm 属性获取对应的 Observer.
    """
    return TensorObserverFactroy.build_observer(variable=variable, config=quant_config)


class TensorQuant(torch.nn.Module):
    def __init__(self, quant_config: TensorQuantizationConfig) -> None:
        """
        PPQ Tensor Quant

        Args:
            quant_config (TensorQuantizationConfig): _description_
            name (str, optional): _description_. Defaults to 'PPQ Quant Stub'.
        """
        self._quant_config = quant_config
        self._delegator = None
        self._batch_observed = 0
        self._observer = Observer(quant_config=quant_config)

    @property
    def delegator(self) -> Callable:
        return self._delegator

    @delegator.setter
    def delegator(self, func: Callable):
        self._delegator = func

    def forward(self, value: torch.Tensor) -> torch.Tensor:
        if self._delegator is not None:
            return self._delegator(value, self._quant_config)
        return QuantFunction(tensor=value, config=self._quant_config)

    def observe(self, value: torch.Tensor):
        self._batch_observed += 1
        self._observer.observe(value)

    def render(self):
        if self._batch_observed == 0:
            raise PermissionError(
                'You have not provide any data to this QuantStub, PPQ can not render its quant config yet.'
            )
        self._observer.render_quantization_config()


class ParameterQuant(TensorQuant):
    def __init__(self, quant_config: TensorQuantizationConfig, parameter: torch.Tensor) -> None:
        if not isinstance(parameter, torch.Tensor):
            raise TypeError(f'Expect a torch.Tensor here. However {type(parameter)} was given.')

        super().__init__(quant_config)
        self.observe(parameter)
        self.render()


def LinearQuantizationConfig(
    symmetrical: bool = True,
    dynamic: bool = False,
    power_of_2: bool = False,
    channel_axis: int = None,
    quant_min: int = -128,
    quant_max: int = 127,
    num_of_bits=8,
    calibration: str = 'minmax',
    rounding: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN,
) -> TensorQuantizationConfig:
    """
    Create a Linear Quantization Config.

    创建线性量化配置信息。
    """

    sym = QuantizationProperty.SYMMETRICAL if symmetrical else QuantizationProperty.ASYMMETRICAL
    dyn = QuantizationProperty.DYNAMIC if dynamic else 0
    pw2 = QuantizationProperty.POWER_OF_2 if power_of_2 else 0
    chn = QuantizationProperty.PER_TENSOR if channel_axis is None else QuantizationProperty.PER_CHANNEL

    return TensorQuantizationConfig(
        policy=QuantizationPolicy(sym + dyn + pw2 + chn + QuantizationProperty.LINEAR),
        rounding=rounding,
        num_of_bits=num_of_bits,
        quant_min=quant_min,
        quant_max=quant_max,
        observer_algorithm=calibration,
        channel_axis=channel_axis,
    )


def FloatingQuantizationConfig(
    symmetrical: bool = True,
    power_of_2: bool = True,
    channel_axis: int = None,
    quant_min: float = -448.0,
    quant_max: float = 448.0,
    exponent: int = 4,
    mantissa: int = 3,
    calibration: str = 'constant',
    rounding: RoundingPolicy = RoundingPolicy.ROUND_HALF_EVEN,
) -> TensorQuantizationConfig:
    """
    Create a Floating Quantization Config.

    创建浮点量化配置信息。
    """

    sym = QuantizationProperty.SYMMETRICAL if symmetrical else QuantizationProperty.ASYMMETRICAL
    pw2 = QuantizationProperty.POWER_OF_2 if power_of_2 else 0
    chn = QuantizationProperty.PER_TENSOR if channel_axis is None else QuantizationProperty.PER_CHANNEL

    return TensorQuantizationConfig(
        policy=QuantizationPolicy(sym + pw2 + chn + QuantizationProperty.FLOATING),
        rounding=rounding,
        num_of_bits=exponent + mantissa + 1,
        exponent_bits=exponent,
        quant_min=quant_min,
        quant_max=quant_max,
        observer_algorithm=calibration,
    )


def Dispatcher(graph: BaseGraph, method: str = 'conservative') -> GraphDispatcher:
    """
    Get a Graph Dispatcher.

    获取一个指定的调度器。
    """
    if method not in DISPATCHER_TABLE:
        raise KeyError(f'Can not find a dispatcher named {method}, check your input again.')
    return DISPATCHER_TABLE[method](graph)


def OperationForwardFunction(optype: str, platform: TargetPlatform = TargetPlatform.FP32) -> Callable:
    """
    Get an Operation forward function. Same Op are allows to have different forward function on different platform,
    to get a default forward function, use platform=TargetPlatform.FP32.

    获取一个算子前向传播执行函数。在 PPQ 中，相同的算子可以在不同的平台上注册成不同的执行逻辑，
    使用 platform = TargetPlatform.FP32 来获取默认执行逻辑。
    """
    if not isinstance(platform, TargetPlatform):
        raise TypeError('Wrong parameter type for invoking this function.')
    if optype not in OPERATION_FORWARD_TABLE[platform]:
        raise KeyError(
            f'Can not find a forward function related to optype {optype}({platform.name}), Register it first.'
        )
    return OPERATION_FORWARD_TABLE[platform][optype]


def Exporter(platform: TargetPlatform) -> GraphExporter:
    """
    Get an network Exporter.

    获取一个网络导出器。
    """
    if not isinstance(platform, TargetPlatform):
        raise TypeError('Wrong parameter type for invoking this function.')
    if platform not in __EXPORTERS__:
        raise KeyError(f'Platfrom {platform.name} has no related exporter, register a exporter for it first.')
    return __EXPORTERS__[platform]()


def Parser(framework: NetworkFramework) -> GraphExporter:
    """
    Get an network Parser.

    获取一个网络解析器。
    """
    if not isinstance(framework, NetworkFramework):
        raise TypeError('Parameter framework has invalid type, Check your parameter again.')
    if framework not in __PARSERS__:
        raise KeyError(f'Requiring framework {framework} does not support parsing now.')
    return __PARSERS__[framework]()


__all__ = [
    'Parser',
    'Exporter',
    'OperationForwardFunction',
    'Dispatcher',
    'FloatingQuantizationConfig',
    'LinearQuantizationConfig',
    'TensorQuant',
    'Quantizer',
    'Observer',
    'Pipeline',
    'QuantFunction',
]



from typing import Union

import torch

from esp_ppq.api.setting import QuantizationSetting
from esp_ppq.core import (
    OperationQuantizationConfig,
    QuantizationPolicy,
    QuantizationProperty,
    QuantizationStates,
    RoundingPolicy,
    TargetPlatform,
)
from esp_ppq.executor import BaseGraphExecutor
from esp_ppq.IR import BaseGraph, Operation
from esp_ppq.quantization.optim import NCNNFormatGemmPass, QuantizationOptimizationPipeline

from .base import BaseQuantizer


class NCNNQuantizer(BaseQuantizer):
    def __init__(self, graph: BaseGraph) -> Union[torch.Tensor, list, dict]:
        super().__init__(graph=graph)
        self._num_of_bits = 8
        self._quant_min = -127
        self._quant_max = +127

    def build_prequant_pipeline(
        self, setting: QuantizationSetting, executor: BaseGraphExecutor
    ) -> QuantizationOptimizationPipeline:
        pipeline = super().build_prequant_pipeline(setting, executor)
        pipeline.append_optimization_to_pipeline(NCNNFormatGemmPass(), at_front=True)
        return pipeline

    def init_quantize_config(self, operation: Operation) -> OperationQuantizationConfig:
        base_quant_config = self.create_default_quant_config(
            policy=self.quantize_policy,
            rounding=self.rounding_policy,
            op=operation,
            num_of_bits=self._num_of_bits,
            exponent_bits=0,
            quant_max=self._quant_max,
            quant_min=self._quant_min,
            observer_algorithm='percentile',
        )

        if operation.type in {'Conv', 'Gemm'}:
            assert operation.num_of_input > 0, 'Seems you got a Computing layer with no parameters.'

            if operation.type == 'Conv':
                conv_weight_config = base_quant_config.input_quantization_config[1]
                conv_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                conv_weight_config.channel_axis = 0
                conv_weight_config.observer_algorithm = 'minmax'

                group = operation.attributes.get('group', 1)
                dilations = operation.attributes.get('dilations', [1, 1])
                strides = operation.attributes.get('strides', [1, 1])
                kernel_shape = operation.attributes.get('kernel_shape')
                if (
                    group == 1
                    and all([i == 1 for i in dilations])
                    and all([j == 1 for j in strides])
                    and all([k == 3 for k in kernel_shape])
                ):
                    base_quant_config.input_quantization_config[1].num_of_bits = 6
                    base_quant_config.input_quantization_config[1].quant_max = +31
                    base_quant_config.input_quantization_config[1].quant_min = -31

            elif operation.type == 'Gemm':
                assert (
                    operation.attributes.get('transB', 0)
                    and operation.attributes.get('alpha', 1.0) == 1.0
                    and operation.attributes.get('beta', 1.0) == 1.0
                )
                gemm_weight_config = base_quant_config.input_quantization_config[1]
                gemm_weight_config.policy = QuantizationPolicy(
                    QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_CHANNEL
                )
                gemm_weight_config.channel_axis = 0
                gemm_weight_config.observer_algorithm = 'minmax'

            # if operation has bias
            if operation.num_of_input > 2:
                bias_config = base_quant_config.input_quantization_config[-1]
                bias_config.state = QuantizationStates.FP32

            base_quant_config.output_quantization_config[0].state = QuantizationStates.FP32
        return base_quant_config

    @property
    def target_platform(self) -> TargetPlatform:
        return TargetPlatform.NCNN_INT8

    @property
    def default_platform(self) -> TargetPlatform:
        return TargetPlatform.FP32

    @property
    def quant_operation_types(self) -> set:
        return {'Conv', 'Gemm'}

    @property
    def quantize_policy(self) -> QuantizationPolicy:
        return QuantizationPolicy(
            QuantizationProperty.SYMMETRICAL + QuantizationProperty.LINEAR + QuantizationProperty.PER_TENSOR
        )

    @property
    def rounding_policy(self) -> RoundingPolicy:
        return RoundingPolicy.ROUND_HALF_FAR_FORM_ZERO

    @property
    def activation_fusion_types(self) -> set:
        """
        ncnn 只需要输入定点 不需要考虑激活函数融合

        Returns:
            set: _description_
        """
        return {}



from esp_ppq.core.defs import ppq_legacy
from esp_ppq.log import NaiveLogger

logger = NaiveLogger.get_logger('PPQ')


# attribute checker and preprocess
def process_attribute(attr, input_shape, kernel_shape=None, op_type=None):
    # ASSUME input is 2D
    # assert len(input_shape) == 2
    # Get default attr value
    auto_pad = attr.get('auto_pad', 'NOTSET')
    strides = attr.get('strides', [1, 1])
    dilations = attr.get('dilations', [1, 1])
    kernels = attr.get('kernel_shape', kernel_shape)
    pad_needed = None

    if op_type == 'ConvTranspose' and 'output_shape' in attr:
        output_shape = attr['output_shape']
        out_pad = [0, 1] if output_shape % 2 != 0 else [0, 0]
        pad_needed = [
            (input_shape[i] - 1) * strides[i] + dilations[i] * (kernels[i] - 1) + 1 + out_pad[i] - output_shape[i]
            for i in range(len(input_shape))
        ]

    if auto_pad != 'NOTSET':
        if 'pads' in attr:
            logger.warning('auto_pad is conflict with pads attribute. Use pads here.')
        elif auto_pad == 'VALID':
            attr['pads'] = [0, 0, 0, 0]
        elif auto_pad in ('SAME_UPPER', 'SAME_LOWER'):
            if op_type == 'ConvTranspose':
                # `output_padding` is only used to find output shape, but does not actually add zero-padding to output
                out_pad = attr.get('output_padding', [0, 0])
                output_shape = [input_shape[i] * strides[i] for i in range(len(input_shape))]
                pad_needed = [
                    (input_shape[i] - 1) * strides[i]
                    + dilations[i] * (kernels[i] - 1)
                    + 1
                    + out_pad[i]
                    - output_shape[i]
                    for i in range(len(input_shape))
                ]
            else:
                output_shape = [(input_shape[i] + strides[i] - 1) // strides[i] for i in range(len(input_shape))]
                pad_needed = [
                    (output_shape[i] - 1) * strides[i] + dilations[i] * (kernels[i] - 1) + 1 - input_shape[i]
                    for i in range(len(input_shape))
                ]
        else:
            raise ValueError(f'Invalid auto_pad value {auto_pad}')

    if pad_needed is not None:
        pads = []
        for item in pad_needed:
            pads.append((item if auto_pad == 'SAME_UPPER' else item + 1) // 2)
        # onnx pads format should be as follow [x1_begin, x2_begin...x1_end, x2_end,...]
        pads = pads + [pad_needed[i] - p for i, p in enumerate(pads)]
        attr['pads'] = pads
        # onnx pads attribute cannot be used simultaneously with auto_pad attribute
        attr.pop('auto_pad')


def preprocess_attr(attr, op_type=None):
    processed_attribute = {}
    if 'kernel_shape' in attr and op_type == 'Pooling':
        processed_attribute['kernel_size'] = attr['kernel_shape']
    if 'group' in attr:
        processed_attribute['groups'] = attr['group']
    if 'pads' in attr:
        # Change pads from start-end to torch format
        pads = attr['pads']
        assert len(pads) % 2 == 0
        if len(pads) == 4:
            begin_pad = pads[:2]
            end_pad = pads[2:]
            if begin_pad == end_pad:
                processed_attribute['padding'] = begin_pad
            else:
                raise ValueError('Torch function only support begin_pad == end_pad in layer')
        else:
            processed_attribute['padding'] = pads

    if 'dilations' in attr:
        processed_attribute['dilation'] = attr['dilations']
    if 'strides' in attr:
        processed_attribute['stride'] = attr['strides']
    if 'ceil_mode' in attr:
        processed_attribute['ceil_mode'] = bool(attr['ceil_mode'])
    return processed_attribute



from typing import Iterable, Tuple

import torch
from tqdm import tqdm

from esp_ppq.executor import TorchExecutor
from esp_ppq.IR import BaseGraph, TrainableGraph


class ImageNetTrainer:
    """
    ### Network Trainer for ImageNet Classification Task.

    Member Functions:
    1. epoch(): do one epoch training.
    2. step(): do one step training.
    3. eval(): evaluation on given dataset.
    4. save(): save trained model.
    5. clear(): clear training cache.

    PPQ will create a TrainableGraph on you graph, a wrapper class that
        implements a set of useful functions that enable training. You are recommended
        to edit its code to add new feature on graph.

    Optimizer controls the learning process and determine the parameters values ends up learning,
        you can rewrite the defination of optimizer and learning scheduler in __init__ function.
        Tuning them carefully, as hyperparameters will greatly affects training result.
    """

    def __init__(self, graph: BaseGraph, device: str = 'cuda') -> None:
        self._epoch = 0
        self._step = 0
        self._best_metric = 0
        self._loss_fn = torch.nn.CrossEntropyLoss().to(device)
        self._executor = TorchExecutor(graph, device=device)
        self._training_graph = TrainableGraph(graph)
        self.graph = graph

        for tensor in self._training_graph.parameters():
            tensor.requires_grad = True

        self._optimizer = torch.optim.RAdam(params=self._training_graph.parameters(), lr=3e-5)
        self._lr_scheduler = None

    def epoch(self, dataloader: Iterable) -> float:
        """Do one epoch Training with given dataloader.

        Given dataloader is supposed to be a iterable container of batched data,
            for example it can be a list of [img(torch.Tensor), label(torch.Tensor)].

        If your data has other layout that is not supported by this function,
            then you are supposed to rewrite the logic of epoch function by yourself.
        """
        epoch_loss = 0
        for bidx, batch in enumerate(tqdm(dataloader, desc=f'Epoch {self._epoch}: ', total=len(dataloader))):
            if type(batch) not in {tuple, list}:
                raise TypeError(
                    'Feeding Data is invalid, expect a Tuple or List like [data, label], '
                    f'however {type(batch)} was given. To feed customized data, you have to rewrite '
                    '"epoch" function of PPQ Trainer.'
                )
            if len(batch) != 2:
                raise ValueError(
                    'Unrecognized data format, your dataloader should contains batched data like [data, label]'
                )
            data, label = batch
            data, label = data.cuda(), label.cuda()

            _, loss = self.step(data, label, True)
            epoch_loss += loss

        self._epoch += 1
        print(f'Epoch Loss: {epoch_loss / len(dataloader):.4f}')
        return epoch_loss

    def step(self, data: torch.Tensor, label: torch.Tensor, training: bool) -> Tuple[torch.Tensor, float]:
        """Do one step Training with given data(torch.Tensor) and label(torch.Tensor).

        This one-step-forward function assume that your model have only one input and output variable.

        If the training model has more input or output variable, then you might need to
            rewrite this function by yourself.
        """
        if training:
            pred = self._executor.forward_with_gradient(data)[0]
            loss = self._loss_fn(pred, label)

            loss.backward()
            if self._lr_scheduler is not None:
                self._lr_scheduler.step(epoch=self._epoch)
            self._optimizer.step()
            self._training_graph.zero_grad()

            self._step += 1
            return pred, loss.item()

        else:
            pred = self._executor.forward(data)[0]
            loss = self._loss_fn(pred, label)

            return pred, loss.item()

    def eval(self, dataloader: Iterable) -> float:
        """Do Evaluation process on given dataloader.

        Split your dataset into training and evaluation dataset at first, then
            use eval function to monitor model performance on evaluation dataset.

        Here are some options to prevent overfitting, which helps improve the model performance.
        1. Train with more data.
        2. Data augmentation.
        3. Addition of noise to the input data.
        """
        total_pred, total_correct = 0, 0
        with torch.no_grad():
            for bidx, batch in enumerate(tqdm(dataloader, desc='Eval: ')):
                if type(batch) not in {tuple, list}:
                    raise TypeError(
                        'Feeding Data is invalid, expect a Tuple or List like [data, label], '
                        f'however {type(batch)} was given. To feed customized data, you have to rewrite '
                        '"eval" function of PPQ Trainer.'
                    )
                if len(batch) != 2:
                    raise ValueError(
                        'Unrecognized data format, your dataloader should contains batched data like [data, label]'
                    )
                data, label = batch
                data, label = data.cuda(), label.cuda()

                pred, _ = self.step(data, label, False)
                pred_label = torch.argmax(pred, dim=-1)
                total_correct += torch.sum(pred_label == label).item()
                total_pred += pred_label.numel()

        print(f'Classification Accuracy: {total_correct / total_pred * 100:.3f}%\n')
        return total_correct / total_pred

    def save(self, file_path: str):
        """Save model to given path.
        Saved model can be read by esp_ppq.api.load_native_model function.
        """
        from esp_ppq.parser import NativeExporter

        exporter = NativeExporter()
        exporter.export(file_path=file_path, graph=self.graph)

    def clear(self):
        """Clear training state."""
        for tensor in self._training_graph.parameters():
            tensor.requires_grad = False
            tensor._grad = None



ONNX_PATH = 'models/yolov5s.v5.onnx'  # 你的模型位置
ENGINE_PATH = 'Output/yolov5s.v5(fp32).engine'  # 生成的 Engine 位置

# -------------------------------------------------------------------
# 打开 trt_infer 看到具体细节，这个文件是 nvidia 的官方实例
# -------------------------------------------------------------------
from trt_infer import EngineBuilder

builder = EngineBuilder()
builder.create_network(ONNX_PATH)
builder.create_engine(engine_path=ENGINE_PATH, precision="fp32")



from typing import Any, Dict, Iterable, List

import onnx
from onnx import helper, numpy_helper

from esp_ppq.core import (
    DEFAULT_OPSET_DOMAIN,
    DEFAULT_OPSET_VERSION,
    GRAPH_OPSET_ATTRIB,
    NetworkFramework,
    is_file_exist,
)
from esp_ppq.IR import BaseGraph, GraphBuilder, Operation, Opset, Variable


class OnnxParser(GraphBuilder):
    def build_variables(
        self,
        graph: BaseGraph,
        graph_inputs: List[str],
        graph_outputs: List[str],
        op_inputs: Dict[str, list],
        op_outputs: Dict[str, list],
    ) -> BaseGraph:
        var_list = []

        for op_name, _ in graph.operations.items():
            for var_name in op_inputs[op_name]:
                var_list.append(var_name)
            for var_name in op_outputs[op_name]:
                var_list.append(var_name)

        # create all variable at once.
        for var_name in set(var_list):
            graph.variables[var_name] = Variable(name=var_name)

        # build graph's input, output variables.
        try:
            for var_name in graph_inputs:
                if var_name not in graph.variables:
                    continue
                graph.inputs[var_name] = graph.variables[var_name]
            for var_name in graph_outputs:
                graph.outputs[var_name] = graph.variables[var_name]
        except KeyError as e:
            raise KeyError('seems you got an input/output variable that is not linked to any operation.')

        # build operation inputs, outputs variables.
        for op in graph.operations.values():
            for var_name in op_inputs[op.name]:
                var = graph.variables[var_name]
                var.dest_ops.append(op)
                op.inputs.append(graph.variables[var_name])
            for var_name in op_outputs[op.name]:
                var = graph.variables[var_name]
                var.source_op = op
                op.outputs.append(graph.variables[var_name])
        return graph

    def initialize_params(self, graph: BaseGraph, initializer: Dict[str, Any]) -> BaseGraph:
        for var in graph.variables.values():
            if var.name in initializer:
                for dest_op in var.dest_ops:
                    assert isinstance(dest_op, Operation)
                    dest_op.parameters.append(var)
                var.value = initializer[var.name]
                var.is_parameter = True
        return graph

    def de_inplace(self, graph: BaseGraph) -> BaseGraph:
        """Remove inplace layer in netdef If the names of bottom and top are same,
        it means the computation of this layer is in place."""

        def new_name(_name):
            if _name == '':
                return ''
            elif _name not in total_write_times or current_write_times:
                return _name
            elif current_write_times[_name] == total_write_times[_name]:
                return _name
            else:
                return f'{_name}_ver{current_write_times[_name]}'

        total_write_times = {}
        for op in graph.operations.values():
            for top in op.outputs:
                total_write_times.setdefault(top._name, 0)
                total_write_times[top._name] += 1

        current_write_times = {}
        for name in graph.inputs.keys():
            total_write_times[name] = 0
            current_write_times[name] = 0

        for op in graph.operations.values():
            for bottom in op.inputs:
                if bottom.is_parameter:
                    continue
                bottom._name = new_name(bottom._name)
            for top in op.outputs:
                current_write_times.setdefault(top._name, 0)
                current_write_times[top._name] += 1
                top._name = new_name(top._name)

    def refine_graph(self, graph: BaseGraph) -> BaseGraph:
        for op in graph.operations.values():
            for key, value in op.attributes.items():
                if isinstance(value, bytes):
                    # Change bytes to string
                    value = value.decode('utf-8')
                if op.type == 'Constant' or op.type == 'ConstantOfShape':
                    # The attribute of 'Constant' node is a value, needs to convert to numpy array
                    value = numpy_helper.to_array(value).copy()
                if op.type == 'Cast' and key == 'to':
                    # The attribute of 'Cast' node is data type (represented in int), need to convert to numpy data type
                    value = helper.tensor_dtype_to_np_dtype[value]
                op.attributes[key] = value

        graph_initializers = []
        for input_var in graph.inputs.values():
            # remove initilizer from graph.inputs
            if input_var.value is not None:
                graph_initializers.append(input_var.name)
        for non_input_var in graph_initializers:
            graph.inputs.pop(non_input_var)
        return graph

    def convert_opsets_to_str(self, opsets: Iterable) -> List[Dict[str, str]]:
        results = []
        for opset in opsets:
            results.append({'domain': opset.domain, 'version': opset.version})
        return results

    def build(self, file_path: str) -> BaseGraph:
        _rand_seed = 0  # used for name generation.
        if not is_file_exist(file_path):
            raise FileNotFoundError(f'file {file_path} does not exist, or it is a directory.')
        model_pb = onnx.load(file_path)
        opsets = model_pb.opset_import

        assert isinstance(model_pb, onnx.ModelProto), (
            f'onnx load failed, only ProtoBuffer object is expected here, while {type(model_pb)} is loaded.'
        )
        graph_pb = model_pb.graph
        graph = BaseGraph(name=graph_pb.name, built_from=NetworkFramework.ONNX)
        graph._detail[GRAPH_OPSET_ATTRIB] = self.convert_opsets_to_str(opsets)
        graph._detail['ir_version'] = model_pb.ir_version

        onnx_import_opset = DEFAULT_OPSET_VERSION
        for opset in graph._detail[GRAPH_OPSET_ATTRIB]:
            if opset['domain'] == DEFAULT_OPSET_DOMAIN or opset['domain'] == '':
                onnx_import_opset = opset['version']
                break

        # a temporary storage for operation's inputs and outputs
        op_inputs_dict, op_outputs_dict = {}, {}
        for node in graph_pb.node:
            op_name = node.name
            if len(op_name) == 0:  # some operation do not have a name, we just generate one.
                op_name = 'generated_name_' + str(_rand_seed)
                _rand_seed += 1

            if op_name in graph.operations:
                raise KeyError(f'Duplicated operation {op_name} was found.')

            graph.operations[op_name] = Operation(
                name=op_name,
                op_type=node.op_type,
                attributes={item.name: helper.get_attribute_value(item) for item in node.attribute},
                opset=Opset(domain=DEFAULT_OPSET_DOMAIN, version=onnx_import_opset),
            )
            op_inputs_dict[op_name] = [var_name for var_name in node.input]
            op_outputs_dict[op_name] = [var_name for var_name in node.output]

        initializer = {}
        for item in graph_pb.initializer:
            init_name = item.name
            value = numpy_helper.to_array(item)
            initializer[init_name] = value

        inputs = [item.name for item in graph_pb.input]
        outputs = [item.name for item in graph_pb.output]
        graph = self.build_variables(
            graph, graph_inputs=inputs, graph_outputs=outputs, op_inputs=op_inputs_dict, op_outputs=op_outputs_dict
        )
        graph = self.initialize_params(graph, initializer)
        self.de_inplace(graph)
        return self.refine_graph(graph)



from typing import Callable, Iterable, List

import torch
from torch.cuda import empty_cache
from tqdm import tqdm

from esp_ppq.core import *
from esp_ppq.executor import BaseGraphExecutor, TorchExecutor
from esp_ppq.IR import BaseGraph, QuantableOperation
from esp_ppq.quantization.algorithm.exprimental import BanditDelegator
from esp_ppq.quantization.algorithm.training import *
from esp_ppq.quantization.measure import torch_snr_error
from esp_ppq.quantization.optim.base import QuantizationOptimizationPass
from esp_ppq.utils.ema import EMARecorder

from .training import TrainingBasedPass


class LearningToCalibPass(TrainingBasedPass):
    """This is an Experimental Pass, do not invoke.

    PPQ Learning Based Calibration Pass
    For int8 quantization, you need to calibrate or estimate the value range,
        i.e, (min, max) of all floating-point tensors in the model.

    Choose value range carefully is really importance procedure during quantization.
        Usually we use methods like MSE, Percentile, KL to solve a good value range
        from prospective view, while this pass offers you another possibility.

    This pass will make all your quantization range as trainable, and learn to quantize
        your network with sampling methods.

    ATTENTION: YOU SHALL USE THIS FUNCTION AFTER ACTIVATIONS HAVE BEEN CORRECTLY CALIBRATED
        SINCE THIS FUNCTION NEEDS A SCALE AND OFFSET AS INITIALIZED VALUE.

    ATTENTION: ONLY CONFIGURATION WITH STATE "ACTIVATED" WILL BE TUNED VIA THIS FUNCTION.
    """

    def __init__(self, method: str = 'e-greedy', calib_act: bool = True, calib_weight: bool = True) -> None:
        self.method = method
        self.calib_act = calib_act
        self.calib_weight = calib_weight
        self.target_step = 7500
        self.e = 0.1
        self.collecting_device = 'cuda'
        self.arms = [1, 0.9, 1.1, 0.7, 1.3]
        # for power-of-2 policy, use bandit like [0.5, 1, 2]
        super().__init__('RL Based Calibration Pass')

    @torch.no_grad()
    def calib_block(
        self,
        quant_inputs: List[torch.Tensor],
        fp32_outputs: List[torch.Tensor],
        executor: TorchExecutor,
        block: TrainableBlock,
        dataloader: Iterable,
        collate_fn: Callable,
    ):
        # create trainable delegators for each parameter.
        delegators = []
        for operation in block.rps:
            if isinstance(operation, QuantableOperation):
                for cfg, var in operation.config_with_variable:
                    if cfg.state == QuantizationStates.ACTIVATED:
                        delegators.append(BanditDelegator(arms=self.arms, config=cfg))
        delegators = [d for d in delegators if isinstance(d, BanditDelegator)]
        dataset = RandomMemDataset(data=[[qt, fp] for qt, fp in zip(quant_inputs, fp32_outputs)])
        device = executor._executing_context.executing_device
        loss_ema = EMARecorder(beta=0.98)
        output_var = block.ep.outputs[0]
        input_var = block.sp.inputs[0]

        for delegator in delegators:
            executor.register_quantize_delegate(config=delegator.config, delegator=delegator)

        cur_iter = 0
        with tqdm(total=self.target_step) as t:
            while cur_iter < self.target_step:
                qt_input, fp_output = dataset.pop()
                qt_input, fp_output = qt_input.to(device), fp_output.to(device)

                qt_output = executor.partial_graph_forward(
                    operations=block.rps, feed_dict={input_var.name: qt_input}, output_names=[output_var.name]
                )[0]

                loss = torch_snr_error(y_pred=qt_output, y_real=fp_output).item()
                for delegator in delegators:
                    delegator.mark(1 - loss)
                loss_ema.push(loss)

                cur_iter += 1
                if cur_iter % 50 == 0:
                    t.set_description(desc=f'Block [{self._bidx + 1}/{self._num_of_blocks}]')
                    t.set_postfix(loss=loss_ema.pop())
                    t.update(50)

        for delegator in delegators:
            executor.remove_quantize_delegate(config=delegator.config)
            delegator.finalize()

        if not self.check(executor=executor, dataloader=dataloader, collate_fn=collate_fn):
            for delegator in delegators:
                delegator.withdraw()

    def collect_training_data(
        self, output_name: str, dataloader: Iterable, executor: BaseGraphExecutor, collate_fn: Callable
    ) -> List[List[torch.Tensor]]:
        output_collector = []
        for data in dataloader:
            if collate_fn is not None:
                data = collate_fn(data)
            [output] = executor.forward(data, output_names=[output_name])
            output_collector.append(output.to(self.collecting_device))
        return output_collector

    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, executor: TorchExecutor, collate_fn: Callable, **kwargs
    ) -> None:
        block_builder = BlockBuilder(graph=graph, topo_order=executor._executing_order)

        # check if there is any baked value inside your graph
        for operation in graph.operations.values():
            if isinstance(operation, QuantableOperation):
                for cfg, var in operation.config_with_variable:
                    if cfg.state in {QuantizationStates.BAKED, QuantizationStates.PASSIVE_BAKED}:
                        raise PermissionError(
                            'Can not apply advanced optimization pass when weight value is baked. '
                            f'Variable {var.name} has a baked value.'
                        )

        # build all blocks, drop overlapped layers.
        blocks, visited = [], set()
        for op in graph.operations.values():
            if op in visited:
                continue
            block = block_builder.build(op, limit=OPTIM_ADVOPT_GRAPH_MAXDEPTH)

            # PATCH 20220317 drop block that has no computing op.
            if all([rp.is_computing_op == False for rp in block.rps]):
                continue
            if block.sp.is_computing_op == False:
                continue

            for rp in block.rps:
                visited.add(rp)
            blocks.append(block)

        self.initialize_checkpoints(graph=graph, executor=executor, dataloader=dataloader, collate_fn=collate_fn)

        block_builder = BlockBuilder(graph=graph, topo_order=executor._executing_order)
        for bidx, block in enumerate(blocks):
            self._bidx, self._num_of_blocks = bidx, len(blocks)
            assert isinstance(block, TrainableBlock)

            end_op = block.ep
            block_input = block.sp.inputs[0]
            block_output = end_op.outputs[0]

            # dequantize prefix operations and block operations
            for op in graph.operations.values():
                if isinstance(op, QuantableOperation):
                    op.dequantize()
                    # can not use dequantize_immediately cause weight has been changed.
                    # self.dequantize_immediately(op)

            fp32_outputs = self.collect_training_data(
                output_name=block_output.name, dataloader=dataloader, executor=executor, collate_fn=collate_fn
            )

            # quantize prefix operations and block operations
            for op in graph.operations.values():
                if isinstance(op, QuantableOperation):
                    op.restore_quantize_state()

            quant_inputs = self.collect_training_data(
                output_name=block_input.name, dataloader=dataloader, executor=executor, collate_fn=collate_fn
            )

            # start training, solve the best parameters
            self.calib_block(
                quant_inputs=quant_inputs,
                fp32_outputs=fp32_outputs,
                executor=executor,
                block=block,
                dataloader=dataloader,
                collate_fn=collate_fn,
            )

            # empty cache.
            fp32_outputs.clear()
            quant_inputs.clear()
            empty_cache()


class MatrixFactorizationPass(QuantizationOptimizationPass):
    """Use Matrix Farctorization to minimize quantization error. This pass will
    split a computing layer with 2 sub layers.

        before split:  WX + b = Y
        after split:   B(AX) + b = Y

        Where W = BA

    However i do not know how to minimize quant loss until now.

    Args:
        QuantizationOptimizationPass ([type]): [description]

    Returns:
        [type]: [description]
    """

    def __init__(self, interested_layers: List[str], method: str = 'training', name: str = 'SVD Split Pass') -> None:
        self.interested_layers = interested_layers
        self.method = method
        super().__init__(name=name)

    @empty_ppq_cache
    def train_for_factorization(
        self, w: torch.Tensor, penalty=0.1, executing_device: str = 'cuda', max_iter: int = 100000
    ):
        assert w.ndim == 2

        a, b = torch.rand(size=[w.shape[0], w.shape[1]]), torch.rand(size=[w.shape[1], w.shape[1]])

        a = a.to(executing_device)
        b = b.to(executing_device)
        w = w.to(executing_device)

        a.requires_grad = True
        b.requires_grad = True
        w.requires_grad = True

        optimizer = torch.optim.Adam(params=[a, b], lr=1e-3)
        loss_fn = torch.nn.MSELoss()

        last_loss = 1e9
        for _ in tqdm(range(max_iter), 'Training for factorization ...'):
            penalty_loss = (torch.mean(torch.square(a)) + torch.mean(torch.square(b))) * penalty
            loss = loss_fn(w, torch.matmul(a, b)) + penalty_loss
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()

            loss = loss.item()
            if abs(loss - last_loss) < 1e-7:
                break
            else:
                last_loss = loss

        a.requires_grad = False
        b.requires_grad = False
        a._grad = None
        b._grad = None

        return a, b

    def svd_for_factorization(self, w: torch.Tensor):
        assert w.ndim == 2
        u, s, v = torch.svd(w)
        a = torch.matmul(u, torch.diag(torch.sqrt(s)))
        b = torch.matmul(torch.diag(torch.sqrt(s)), v.transpose(0, 1))
        print(a.max(), b.max(), w.max())
        return a, b

    def optimize(self, graph: BaseGraph, dataloader: Iterable, executor: BaseGraphExecutor, **kwargs) -> None:
        splitting_layers = []
        for name in self.interested_layers:
            if name not in graph.operations:
                raise ValueError(f'Can not Split layer {name}, can not find it in current graph.')

        for operation in graph.operations.values():
            if operation.name in self.interested_layers:
                assert operation.type in {'Conv', 'Gemm'}, 'Can not split layer, cause layer type is not support'
                splitting_layers.append(operation)

        for operation in splitting_layers:
            assert isinstance(operation, Operation)
            if operation.type == 'Gemm':
                w = operation.parameters[0].value
                w = w.transpose(0, 1)
                if self.method == 'svd':
                    a, b = self.svd_for_factorization(w)
                elif self.method == 'training':
                    a, b = self.train_for_factorization(w)
                else:
                    raise ValueError(f'Invalid method {self.method}, only support training and svd now.')
                a = a.transpose(0, 1)
                b = b.transpose(0, 1)
            elif operation.type == 'Conv':
                if operation.attributes['kernel_shape'] != [1, 1]:
                    raise PermissionError(f'Can not split layer {operation.name}, cause it kernel shape is not [1, 1]')
                w = operation.parameters[0].value
                assert isinstance(w, torch.Tensor)
                w = w.squeeze(-1).squeeze(-1).transpose(0, 1)
                print(w.shape)
                if self.method == 'svd':
                    a, b = self.svd_for_factorization(w)
                elif self.method == 'training':
                    a, b = self.train_for_factorization(w)
                else:
                    raise ValueError(f'Invalid method {self.method}, only support training and svd now.')
                a = a.transpose(0, 1).unsqueeze(-1).unsqueeze(-1)
                b = b.transpose(0, 1).unsqueeze(-1).unsqueeze(-1)
            else:
                raise TypeError(f'Unsupported operation type {operation.type}.')
            operation.parameters[0].value = a

            # create new operation & dirty work
            attributes = {}
            if operation.type == 'Conv':
                attributes['kernel_shape'] = [1, 1]
                attributes['pads'] = [0, 0, 0, 0]
                attributes['strides'] = [1, 1]
                attributes['dilations'] = [1, 1]
                attributes['group'] = 1

            if operation.type == 'Gemm':
                attributes['alpha'] = 1
                attributes['beta'] = 1
                attributes['transB'] = 1

            splitted = Operation(
                name=operation.name + '_splited',
                op_type=operation.type,
                attributes=attributes,
                platform=operation.platform,
            )

            graph.insert_op_on_var(inserting_op=splitted, var=operation.outputs[0].name)
            if operation.outputs[0].name in graph.outputs:
                graph.outputs.pop(operation.outputs[0].name)
                graph.outputs[splitted.outputs[0].name] = splitted.outputs[0]

            # add weight link
            spilted_w = Variable(name=splitted.name + '_weight', value=b, is_parameter=True)
            graph.append_variable(spilted_w)

            splitted.inputs.append(spilted_w)
            spilted_w.dest_ops.append(splitted)

            # if has bias, relink bias
            if len(operation.parameters) > 1:
                bias_var = operation.parameters[-1]
                bias_var.dest_ops.remove(operation)
                bias_var.dest_ops.append(splitted)
                splitted.inputs.append(bias_var)
                operation.inputs.remove(bias_var)



from .imagenet_util import (
    evaluate_mmlab_module_with_imagenet,
    evaluate_onnx_module_with_imagenet,
    evaluate_ppq_module_with_imagenet,
    evaluate_torch_module_with_imagenet,
    load_imagenet_from_directory,
)



from typing import Iterable

from esp_ppq.core import empty_ppq_cache
from esp_ppq.executor import BaseGraphExecutor
from esp_ppq.IR import BaseGraph, QuantableOperation
from esp_ppq.quantization.qfunction import PPQuantFunction

from .base import QuantizationOptimizationPass


class ParameterBakingPass(QuantizationOptimizationPass):
    """ParameterBakingPass is a useful tool for quantization simulation
    acceleration. By default quantizer will bake network parameters once all
    quantization procedures are finished. For a typical Convolution layer or
    Gemm layer, which has a non-empty bias tensor, ParameterBakingPass will
    speed up the layer execution by 30%-50%.

    ParameterBakingPass will rewrite layer parameters with their quantized version,
        the quantization procedure will strictly follow layer quantization configuration.
    Once the quantization process finished, this pass will change all parameter quantization configuration states
        to QuantizationStates.BAKED.

    State QuantizationStates.BAKED indicates corresponding tensor has been pre-quantized and its value
        can be used without further quantization, executor will directly use a baked value during execution.

    ATTENTION: value is baked inplace, so to say it will rewrite all network parameters.
    ATTENTION: For platforms using int32 accumulator, a float32 bias tensor might lose precision
        during the simulation. If you want PPQ simulator to have a consistent result with hardware, it is
        highly-recommended to calling ParameterBakingPass before deployment, baking procedure will limit bias
        precision to 23 bits (float32 only has 23 fraction bits).
    Args:
        quantize_function (BaseQuantFunction): a BaseQuantFunction instance to quantize all parameters.
    """

    def __init__(self) -> None:
        super().__init__(name='PPQ Parameter Baking Pass')
        self._quantize_function = PPQuantFunction

    @empty_ppq_cache
    def optimize(self, graph: BaseGraph, **kwargs) -> None:
        for _, operation in graph.operations.items():
            if not isinstance(operation, QuantableOperation):
                continue
            operation.baking_parameters(self._quantize_function)



from abc import ABCMeta, abstractmethod, abstractproperty
from typing import Any, Callable, List, Union

from .base.command import GraphCommand, GraphCommandType
from .base.graph import BaseGraph


class GraphCommandProcessor(Callable, metaclass=ABCMeta):
    def __init__(self, graph_or_processor: Union[BaseGraph, Callable]) -> None:
        """GraphCommandProcessor 是用于处理图上相关操作的抽象基类.

            我们使用指令-责任链模式处理 PPQ 计算图的相关操作，具体来说：

                所有图上相关操作都由一个 GraphCommand 对象进行封装，
                这一对象封装了操作的类型和必要参数

                同时我们设计了 GraphCommandProcessor 类用于接收并处理对应的 GraphCommand

                GraphCommandProcessor 被设计为责任链模式，当接收到无法处理的 GraphCommand 时
                将把无法识别 GraphCommand 传递给责任链上的下一任 GraphCommandProcessor
                直到 GraphCommand 被处理，或抛出无法处理的异常

                当你实现一个新的 GraphCommandProcessor 时，需要实现其中的方法 _acceptable_command_types，
                该方法返回了所有可以被识别的 GraphCommand 类型，同时在 _process 的逻辑中对 GraphCommand 的请求进行处理

                这两个方法被设计为私有的，这意味着你不能单独访问责任链中的独立 GraphCommandProcessor，
                只能够通过责任链的方式发起请求

                如果在责任链中有多个可以处理同一类请求的 GraphCommandProcessor，
                只有最先接触到 GraphCommand 的 GraphCommandProcessor将会被调用

                GraphCommandProcessor 将按照自定义逻辑解析 GraphCommand，
                在 BaseGraph 做出相应处理并给出结果，实现方法请参考 RunnableGraph

            GraphCommandProcessor is an abstract class for manipulating the graph

            We use Command-Responsibility Chain to process operations on the computational graph:

                all graph-related operations are encapsulated by GraphCommand objects, which contain
                the operation type and necessary parameters

                we design GraphCommandProcessor to process corresponding GraphCommand

                GraphCommandProcessor follows responsibility chain convention, a processor will pass
                the unrecogonized GraphCommand to the next processor in the chain, the GraphCommand will
                finally be executed when meeting the corresponding processor. An exception will be thrown
                if no processor in the current chain is able to deal with the given command

                When you implement a new GraphCommandProcessor, you should implement its inner method
                _acceptable_command_types, it returns all available GraphCommand types for execution
                and you should implement corresponding execution details in _process

                You are not supposed to visit single GraphCommandProcessor in the chain, and if there are
                multiple processors which could deal with same Command type in the chain, the first processor
                receiving the command will execute

                GraphCommandProcessor follows its predefined logic to parse GraphCommand, execute on the graph
                correspondingly and return the final result

        Args:
            graph_or_processor (BaseGraph, Callable): 被处理的图对象，可以是 BaseGraph 或者 GraphCommandProcessor
                如果是 GraphCommandProcessor 对象，则自动将 self 与 graph 链接成链
                the graph being executed or previous GraphCommandProcessor in the chain
        """
        if isinstance(graph_or_processor, GraphCommandProcessor):
            self._next_command_processor = graph_or_processor
            self._graph = graph_or_processor._graph
        else:
            self._next_command_processor = None
            self._graph = graph_or_processor

    @property
    @abstractproperty
    def _acceptable_command_types(self) -> List[GraphCommandType]:
        """Subclass of GraphCommandProcessor must give an implementation of
        this function.

            Return all acceptable GraphCommandTypes in a list as result.
            something like:
                return [
                    GraphCommandType.DEPLOY_TO_CPU,
                    GraphCommandType.DEPLOY_TO_CUDA,
                    GraphCommandType.DEPLOY_TO_NUMPY
                ]

        Returns:
            List[GraphCommandType]: all acceptable GraphCommandTypes
        """
        raise NotImplementedError(
            'Oh, seems you forgot to implement GraphCommandProcessor._acceptable_command_types function'
        )

    def __call__(self, command: GraphCommand) -> Any:
        """Invoking interface of GraphCommandProcessor responsibility chain.
        All processors within the chain shall be invoked by this function one
        be one, until there is a processor claim to accept input command
        object, the entire processing of responsibility chain ends then.

            invoke a GraphCommandProcessor chain like that:

                _ = GraphCommandProcessor(graph, next_command_processor=None)
                _ = GraphCommandProcessor(graph, next_command_processor=_)
                command_processor = GraphCommandProcessor(graph, next_command_processor=_)

                command = GraphCommand(GraphCommandType.DEPLOY_TO_CUDA)
                command_processor(command)

            All three GraphCommandProcessor will then be called one by one

            Never attempt to use function like _(command) in above case,
            all responsibility chains should only be called by its head.

        Args:
            command (GraphCommand): An acceptable GraphCommand object
            if an improper GraphCommand is given, it will incurs ValueError at end.

        Raises:
            ValueError: raise when there is no suitable processor for your command.

        Returns:
            Any: processor will decide what is it result.
        """

        if not isinstance(command, GraphCommand):
            raise ValueError(f'command should be an instance of GraphCommand, {type(command)} received yet.')

        if command.command_type in self._acceptable_command_types():
            return self.process(command)

        elif self._next_command_processor is not None:
            self._next_command_processor(command)

        else:
            raise ValueError(
                f'Command Type {command.command_type} is not acceptable in this graph, '
                'please make sure you have added proper command processor into processing chain.\n'
                'For more information, you may refer to esp_ppq.IR.graph.GraphCommandType'
            )

    def acceptable_command_types(self) -> List[GraphCommandType]:
        """Return all acceptable command types of current chain. Notice there
        might be duplicated types.

        Returns:
            List[GraphCommandType]: all acceptable command types
        """
        my_types = self._acceptable_command_types()
        assert isinstance(my_types, list), (
            'GraphCommandProcessor.__acceptable_command_types must return a list of GraphCommandType'
        )

        if self._next_command_processor is not None:
            other_types = self._next_command_processor.acceptable_command_types()
        return my_types.extend(other_types)

    @abstractmethod
    def process(self, command: GraphCommand) -> Any:
        """Subclass of GraphCommandProcessor must give an implementation of
        this function.

            Process received GraphCommand instance and give result(if there is any)

        Args:
            command (GraphCommand): input command object.

        Returns:
            Any: any result is fine.
        """
        raise NotImplementedError(
            'Oh, seems you forgot to implement GraphCommandProcessor.process function(or may be forgot to return?)'
        )

    @property
    def graph(self) -> BaseGraph:
        return self._graph

    def __str__(self) -> str:
        return f'GraphCommandProcessor {self.__hash__}'


class DefaultGraphProcessor(GraphCommandProcessor):
    def _acceptable_command_types(self) -> List[GraphCommandType]:
        return []

    def process(self, command: GraphCommand) -> Any:
        pass



from typing import Callable, Iterable

import torch
import torchvision

from esp_ppq import TargetPlatform
from esp_ppq.api import ENABLE_CUDA_KERNEL, QuantizationSettingFactory, quantize_torch_model
from esp_ppq.core import QuantizationStates
from esp_ppq.executor.torch import TorchExecutor
from esp_ppq.IR.quantize import QuantableOperation

# ------------------------------------------------------------
# 在这个例子中，我们将向你介绍如何自定义量化优化过程，以及如何手动调用优化过程
# ------------------------------------------------------------

BATCHSIZE = 32
INPUT_SHAPE = [BATCHSIZE, 3, 224, 224]
DEVICE = 'cuda'
PLATFORM = TargetPlatform.TRT_INT8


# ------------------------------------------------------------
# 和往常一样，我们要创建 calibration 数据，以及加载模型
# ------------------------------------------------------------
def load_calibration_dataset() -> Iterable:
    return [torch.rand(size=INPUT_SHAPE) for _ in range(32)]


CALIBRATION = load_calibration_dataset()


def collate_fn(batch: torch.Tensor) -> torch.Tensor:
    return batch.to(DEVICE)


model = torchvision.models.mobilenet.mobilenet_v2(pretrained=True)
model = model.to(DEVICE)

# ------------------------------------------------------------
# 下面，我们将向你展示如何不借助 QSetting 来自定义优化过程
# QSetting 中包含了 PPQ 官方量化过程的配置参数，你可以借助它来调用所有系统内置优化过程
# 但如果你设计了新的优化过程，你将必须在合适的时机手动启动他们
# ------------------------------------------------------------
QSetting = QuantizationSettingFactory.default_setting()
# 不要进行 Parameter Baking 操作，一旦 Parameter 完成 Baking，后续任何对于参数的修改都是不被允许的
# 你可以设置 baking_parameter = True 并再次执行这个脚本，PPQ 系统会拒绝后续修改 scale 的请求
QSetting.quantize_parameter_setting.baking_parameter = False

# ------------------------------------------------------------
# 定义我们自己的优化过程，继承 QuantizationOptimizationPass 基类，实现 optimize 接口
# 在 optimize 接口函数中，你可以修改图的属性从而实现特定目的
# 在这个例子中，我们将图中所有卷积的输入 scale 变换为原来的两倍
# 同时，我们解除最后一个 Gemm 的输入量化
# ------------------------------------------------------------
from esp_ppq import BaseGraph, QuantizationOptimizationPass


class MyOptim(QuantizationOptimizationPass):
    def optimize(
        self, graph: BaseGraph, dataloader: Iterable, collate_fn: Callable, executor: TorchExecutor, **kwargs
    ) -> None:
        # graph.operations 是一个包含了图中所有 op 的字典
        for name, op in graph.operations.items():
            # 从图中找出所有已经量化的卷积算子
            # 对于你的网络而言，并非所有算子最终都会被量化，他们会受到 调度策略 和 Quantizer策略 的双重限制
            # 因此我们要使用 isinstance(op, QuantableOperation) 来判断它是否是一个量化的算子
            if op.type == 'Conv' and isinstance(op, QuantableOperation):
                # 对于卷积算子，它可能有 2-3 个输入，其中第二个输入为权重，第三个输入为 bias
                # 我们修改权重量化信息的 scale
                op.input_quant_config[1].scale *= 2
                print(f'Input scale of Op {name} has been enlarged.')

            # 我们接下来解除 Gemm 的量化，在这里 mobilenet_v2 网络只有一个 Gemm 层
            # 所以我们将所有遇到的 Gemm 的层全部解除量化
            if op.type == 'Gemm' and isinstance(op, QuantableOperation):
                # config_with_variable 接口将返回量化算子的所有量化信息————包括输入与输出
                for cfg, _ in op.config_with_variable:
                    # 在 PPQ 中有许多方法可以切换算子的量化状态
                    # 将量化状态直接设置为 FP32，即解除了算子的量化
                    cfg.state = QuantizationStates.FP32

                # 也可以直接调用算子的 dequantize 方法
                # op.dequantize()


# ------------------------------------------------------------
# 如果你使用 ENABLE_CUDA_KERNEL 方法
# PPQ 将会尝试编译自定义的高性能量化算子，这一过程需要编译环境的支持
# 如果你在编译过程中发生错误，你可以删除此处对于 ENABLE_CUDA_KERNEL 方法的调用
# 这将显著降低 PPQ 的运算速度；但即使你无法编译这些算子，你仍然可以使用 pytorch 的 gpu 算子完成量化
# ------------------------------------------------------------
with ENABLE_CUDA_KERNEL():
    quantized = quantize_torch_model(
        model=model,
        calib_dataloader=CALIBRATION,
        calib_steps=32,
        input_shape=INPUT_SHAPE,
        setting=QSetting,
        collate_fn=collate_fn,
        platform=PLATFORM,
        onnx_export_file='Output/model.onnx',
        device=DEVICE,
        verbose=0,
    )

    # ------------------------------------------------------------
    # 在完成量化流程之后，我们调用我们自定义的量化优化过程从而修改量化参数
    # ------------------------------------------------------------
    optim = MyOptim(name='My Optimization Procedure')
    optim.optimize(
        graph=quantized,
        dataloader=CALIBRATION,
        collate_fn=INPUT_SHAPE,
        executor=TorchExecutor(quantized, device=DEVICE),
    )
